import argparse
import os
import pickle

import cv2
import depthai as dai
from utils import extract_main_slug, extract_parser, get_input_shape

argparser = argparse.ArgumentParser()
argparser.add_argument("-ip", help="IP address of the device", default="")
argparser.add_argument(
    "-img", "--img_path", help="Path to the input image", required=True, type=str
)
argparser.add_argument(
    "-m", "--model", help="Model from HubAI", required=True, type=str
)

args = argparser.parse_args()
IP_mxid = args.ip
img_path = args.img_path
model: str = args.model

device = dai.Device(dai.DeviceInfo(IP_mxid))
device_platform = device.getPlatform().name

# Get the model from the HubAI
model_description = dai.NNModelDescription(model=model, platform=device_platform)
archive_path = dai.getModelFromZoo(model_description)
nn_archive = dai.NNArchive(archive_path)

input_size = get_input_shape(nn_archive)
input_width = input_size[0]
input_height = input_size[1]
parser = extract_parser(nn_archive)


class HostImage(dai.node.ThreadedHostNode):
    def __init__(self):
        dai.node.ThreadedHostNode.__init__(self)
        self.out = self.createOutput()

    def run(self):
        frame = cv2.imread(img_path)
        if frame is None:
            raise FileNotFoundError(f"Cannot read image from {img_path}")
        else:
            frame = cv2.resize(frame, (input_width, input_height))
            if device_platform == "RVC2":
                frame = frame.transpose(2, 0, 1)

            imgFrame = dai.ImgFrame()

            imgFrame.setFrame(frame)
            imgFrame.setWidth(input_width)
            imgFrame.setHeight(input_height)
            imgFrame.setType(
                dai.ImgFrame.Type.BGR888p
                if device_platform == "RVC2"
                else dai.ImgFrame.Type.BGR888i
            )
            print("input frame HW:", imgFrame.getHeight(), imgFrame.getWidth())

            self.out.send(imgFrame)


with dai.Pipeline(device) as pipeline:
    image_node = pipeline.create(HostImage)

    model_platforms = [platform.name for platform in nn_archive.getSupportedPlatforms()]

    if device_platform not in model_platforms:
        print(f"Model not supported on {device_platform}.")
        device.close()
        exit(5)

    detection_nn = pipeline.create(dai.node.NeuralNetwork)
    detection_nn.setNNArchive(nn_archive)
    image_node.out.link(detection_nn.input)

    nn_queue = detection_nn.out.createOutputQueue()
    pass_queue = detection_nn.passthrough.createOutputQueue()

    pipeline.start()

    while pipeline.isRunning():
        nn_data: dai.NNData = nn_queue.get()
        frame: dai.ImgFrame = pass_queue.get().getCvFrame()
        tensor_names = nn_data.getAllLayerNames()
        tensors = {}
        print("-------------------------------")
        print("Tensor name | Shape | Data type")
        print("-------------------------------")
        for tensor_name in tensor_names:
            tensor = nn_data.getTensor(tensor_name)
            print(tensor_name, tensor.shape, tensor.dtype)
            tensors[tensor_name] = tensor

        if not os.path.exists("nn_datas"):
            os.makedirs("nn_datas")
        if not os.path.exists(f"nn_datas/{parser}"):
            os.makedirs(f"nn_datas/{parser}")
        with open(f"nn_datas/{parser}/{extract_main_slug(model)}.pkl", "wb") as f:
            pickle.dump(tensors, f)
        cv2.imwrite(f"nn_datas/{parser}/{extract_main_slug(model)}.png", frame)

        pipeline.stop()

device.close()



import copy
from typing import List, Tuple, Union

import depthai as dai
import numpy as np
from numpy.typing import NDArray

from depthai_nodes import (
    PRIMARY_COLOR,
    SECONDARY_COLOR,
    TRANSPARENT_PRIMARY_COLOR,
)
from depthai_nodes.logging import get_logger

from .keypoints import Keypoint, Keypoints
from .segmentation import SegmentationMask


class ImgDetectionExtended(dai.Buffer):
    """A class for storing image detections in (x_center, y_center, width, height)
    format with additional angle, label and keypoints.

    Attributes
    ----------
    rotated_rect: dai.RotatedRect
        Rotated rectangle object defined by the center, width, height and angle in degrees.
    confidence: float
        Confidence of the detection.
    label: int
        Label of the detection.
    label_name: str
        The corresponding label name if available.
    keypoints: Keypoints
        Keypoints of the detection. Getter returns list of Keypoint objects and setter accepts Keypoints object.
    """

    def __init__(self):
        """Initializes the ImgDetectionExtended object."""
        super().__init__()
        self._rotated_rect: dai.RotatedRect
        self._confidence: float = -1.0
        self._label: int = -1
        self._label_name: str = ""
        self._keypoints: Keypoints = Keypoints()
        self._logger = get_logger(__name__)

    def copy(self):
        """Creates a new instance of the ImgDetectionExtended class and copies the
        attributes.

        @return: A new instance of the ImgDetectionExtended class.
        @rtype: ImgDetectionExtended
        """
        new_obj = ImgDetectionExtended()
        rectangle = (
            self._rotated_rect.center.x,
            self._rotated_rect.center.y,
            self._rotated_rect.size.width,
            self._rotated_rect.size.height,
            self._rotated_rect.angle,
        )
        new_obj.rotated_rect = copy.deepcopy(rectangle)
        new_obj.confidence = copy.deepcopy(self.confidence)
        new_obj.label = copy.deepcopy(self.label)
        new_obj.label_name = copy.deepcopy(self.label_name)
        new_obj.keypoints = self._keypoints.copy()
        return new_obj

    @property
    def rotated_rect(self) -> dai.RotatedRect:
        """Returns the rotated rectangle representing the bounding box.

        @return: Rotated rectangle object
        @rtype: dai.RotatedRect
        """
        return self._rotated_rect

    @rotated_rect.setter
    def rotated_rect(self, rectangle: Tuple[float, float, float, float, float]):
        """Sets the rotated rectangle of the bounding box.

        @param value: Tuple of (x_center, y_center, width, height, angle).
        @type value: tuple[float, float, float, float, float]
        """
        center = dai.Point2f(rectangle[0], rectangle[1], normalized=True)
        size = dai.Size2f(rectangle[2], rectangle[3], normalized=True)

        self._rotated_rect = dai.RotatedRect(center, size, rectangle[4])

    @property
    def confidence(self) -> float:
        """Returns the confidence of the detection.

        @return: Confidence of the detection.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the detection.

        @param value: Confidence of the detection.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("Confidence must be a float.")
        if value < -0.1 or value > 1.1:
            raise ValueError("Confidence must be between 0 and 1.")
        if not (0 <= value <= 1):
            value = max(0, min(1, value))
            self._logger.info("Confidence value was clipped to [0, 1].")

        self._confidence = value

    @property
    def label(self) -> int:
        """Returns the label of the detection.

        @return: Label of the detection.
        @rtype: int
        """
        return self._label

    @label.setter
    def label(self, value: int):
        """Sets the label of the detection.

        @param value: Label of the detection.
        @type value: int
        @raise TypeError: If value is not an integer.
        """
        if not isinstance(value, int):
            raise TypeError("Label must be an integer.")
        self._label = value

    @property
    def label_name(self) -> str:
        """Returns the label name of the detection.

        @return: Label name of the detection.
        @rtype: str
        """
        return self._label_name

    @label_name.setter
    def label_name(self, value: str):
        """Sets the label name of the detection.

        @param value: Label name of the detection.
        @type value: str
        @raise TypeError: If value is not a string.
        """
        if not isinstance(value, str):
            raise TypeError("Label name must be a string.")
        self._label_name = value

    @property
    def keypoints(
        self,
    ) -> List[Keypoint]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: Keypoints
        """
        return self._keypoints.keypoints

    @keypoints.setter
    def keypoints(
        self,
        value: Keypoints,
    ) -> None:
        """Sets the keypoints.

        @param value: Keypoints object.
        @type value: Keypoints
        @raise TypeError: If value is not a Keypoints object.
        """
        if not isinstance(value, Keypoints):
            raise TypeError("Keypoints must be a Keypoints object.")
        self._keypoints = value


class ImgDetectionsExtended(dai.Buffer):
    """ImgDetectionsExtended class for storing image detections with keypoints.

    Attributes
    ----------
    detections: List[ImgDetectionExtended]
        Image detections with keypoints.
    masks: np.ndarray
        The segmentation masks of the image. All masks are stored in a single numpy array.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self) -> None:
        """Initializes the ImgDetectionsExtended object."""
        super().__init__()
        self._detections: List[ImgDetectionExtended] = []
        self._masks: SegmentationMask = SegmentationMask()
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the ImgDetectionsExtended class and copies the
        attributes.

        @return: A new instance of the ImgDetectionsExtended class.
        @rtype: ImgDetectionsExtended
        """
        new_obj = ImgDetectionsExtended()
        new_obj.detections = [det.copy() for det in self.detections]
        new_obj.masks = self._masks.copy()
        new_obj.transformation = self.transformation
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def detections(self) -> List[ImgDetectionExtended]:
        """Returns the image detections with keypoints.

        @return: List of image detections with keypoints.
        @rtype: List[ImgDetectionExtended]
        """
        return self._detections

    @detections.setter
    def detections(self, value: List[ImgDetectionExtended]):
        """Sets the image detections with keypoints.

        @param value: List of image detections with keypoints.
        @type value: List[ImgDetectionExtended]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type ImgDetectionExtended.
        """
        if not isinstance(value, list):
            raise TypeError("Detections must be a list.")
        if not all(isinstance(detection, ImgDetectionExtended) for detection in value):
            raise TypeError(
                "Detections must be a list of ImgDetectionExtended objects."
            )
        self._detections = value

    @property
    def masks(self) -> NDArray[np.int16]:
        """Returns the segmentation masks stored in a single numpy array.

        @return: Segmentation masks.
        @rtype: SegmentationMask
        """
        return self._masks.mask

    @masks.setter
    def masks(self, value: Union[NDArray[np.int16], SegmentationMask]):
        """Sets the segmentation mask.

        @param value: Segmentation mask.
        @type value: NDArray[np.int8])
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type int8.
        @raise ValueError: If each element is larger or equal to -1.
        """
        if isinstance(value, SegmentationMask):
            self._masks = value
        elif isinstance(value, np.ndarray):
            if not (value.size == 0 or value.ndim == 2):
                raise ValueError("Mask must be 2D.")
            if value.dtype != np.int16:
                raise ValueError("Mask must be an array of int16.")
            if np.any((value < -1)):
                raise ValueError("Mask must be an array values larger or equal to -1.")
            masks_msg = SegmentationMask()
            masks_msg.mask = value
            self._masks = masks_msg
        else:
            raise TypeError("Mask must be a numpy array or a SegmentationMask object.")

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        if transformation is not None:
            assert isinstance(transformation, dai.ImgTransformation)
        self.transformation = transformation

    def getTransformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self.transformation

    def getVisualizationMessage(self) -> dai.ImgAnnotations:
        img_annotations = dai.ImgAnnotations()
        annotation = dai.ImgAnnotation()
        transformation = self.transformation
        w, h = transformation.getSize()

        for detection in self.detections:
            detection: ImgDetectionExtended = detection
            rotated_rect = detection.rotated_rect
            rotated_rect = rotated_rect.denormalize(w, h)
            points = rotated_rect.getPoints()
            points = [dai.Point2f(point.x / w, point.y / h) for point in points]
            pointsAnnotation = dai.PointsAnnotation()
            pointsAnnotation.type = dai.PointsAnnotationType.LINE_LOOP
            pointsAnnotation.points = dai.VectorPoint2f(points)
            pointsAnnotation.outlineColor = PRIMARY_COLOR
            pointsAnnotation.fillColor = TRANSPARENT_PRIMARY_COLOR
            pointsAnnotation.thickness = 1.0
            annotation.points.append(pointsAnnotation)

            text = dai.TextAnnotation()
            text.position = points[0]
            font_size = h / 30
            text.position.y += font_size / h
            text.position.x += 1 / w
            text.position.y += 1 / h
            text.text = f"{detection.label_name} {int(detection.confidence * 100)}%"
            text.fontSize = font_size
            text.textColor = SECONDARY_COLOR
            annotation.texts.append(text)

            if len(detection.keypoints) > 0:
                keypoints = [
                    dai.Point2f(keypoint.x, keypoint.y)
                    for keypoint in detection.keypoints
                ]
                keypointAnnotation = dai.PointsAnnotation()
                keypointAnnotation.type = dai.PointsAnnotationType.POINTS
                keypointAnnotation.points = dai.VectorPoint2f(keypoints)
                keypointAnnotation.outlineColor = PRIMARY_COLOR
                keypointAnnotation.fillColor = PRIMARY_COLOR
                keypointAnnotation.thickness = 2
                annotation.points.append(keypointAnnotation)

                if detection._keypoints.edges is not None:
                    for edge in detection._keypoints.edges:
                        skeletonAnnotation = dai.PointsAnnotation()
                        skeletonAnnotation.type = dai.PointsAnnotationType.LINE_STRIP
                        pt1 = keypoints[edge[0]]
                        pt2 = keypoints[edge[1]]
                        skeletonAnnotation.points = dai.VectorPoint2f(
                            [dai.Point2f(pt1.x, pt1.y), dai.Point2f(pt2.x, pt2.y)]
                        )
                        skeletonAnnotation.outlineColor = SECONDARY_COLOR
                        skeletonAnnotation.fillColor = SECONDARY_COLOR
                        skeletonAnnotation.thickness = 1
                        annotation.points.append(skeletonAnnotation)

        img_annotations.annotations.append(annotation)
        img_annotations.setTimestamp(self.getTimestamp())
        return img_annotations



import logging
import os
from enum import Enum
from typing import Optional


class LogLevel(Enum):
    CRITICAL = "CRITICAL"
    DEBUG = "DEBUG"
    ERR = "ERR"
    INFO = "INFO"
    WARN = "WARN"


def get_logger(name: Optional[str] = None) -> logging.Logger:
    logger = logging.getLogger("depthai-nodes")
    if name:
        logger = logger.getChild(name)
    return logger


def setup_logging(level: Optional[str] = None, file: Optional[str] = None):
    """Globally configures logging for depthai_nodes package.

    @type level: str or None
    @param level: Logging level. One of "CRITICAL", "DEBUG", "ERR", "INFO", and "WARN".
        Can be changed using "DEPTHAI_NODES_LEVEL" env variable. If not set defaults to
        "DEPTHAI_LEVEL" if set or "WARN".
    @type file: str or None
    @param file: Path to a file where logs will be saved. If None, logs will not be
        saved. Defaults to None.
    """
    logger = get_logger()
    passed_level = get_log_level(level)
    env_dai_nodes_level = get_log_level(os.environ.get("DEPTHAI_NODES_LEVEL", None))
    env_dai_level = get_log_level(os.environ.get("DEPTHAI_LEVEL", None))

    used_level = passed_level or env_dai_nodes_level or env_dai_level
    if not used_level:
        used_level = LogLevel.WARN

    format = (
        file_format
    ) = "%(asctime)s [depthai-nodes] [%(levelname)s] [%(name)s] %(message)s"
    datefmt = "[%Y-%m-%d %H:%M:%S]"

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(format, datefmt=datefmt))
    logger.addHandler(console_handler)

    if file is not None:
        file_handler = logging.FileHandler(file)
        file_handler.setFormatter(logging.Formatter(file_format, datefmt=datefmt))
        logger.addHandler(file_handler)

    logger.setLevel(used_level.value)
    logger.info(f"Using log level: {used_level}")


def get_log_level(level_str: Optional[str]) -> Optional[LogLevel]:
    try:
        if level_str is None:
            return None

        level_str = level_str.upper()  # type: ignore
        # mapping some DepthAI specific levels to python logger levels
        if level_str in "OFF":
            level_str = "WARN"
        elif level_str == "TRACE":
            level_str = "INFO"
        return LogLevel(level_str)
    except ValueError as e:
        raise e



from abc import ABCMeta

import depthai as dai

from depthai_nodes.logging import get_logger

HostNodeMeta = type(dai.node.HostNode)  # metaclass of dai.node.HostNode


class CombinedMeta(ABCMeta, HostNodeMeta):
    pass


class BaseHostNode(dai.node.HostNode, metaclass=CombinedMeta):
    """An abstract base class for host nodes.

    Designed to encapsulate and abstract the configuration of platform-specific
    attributes, providing a clean and consistent interface for derived classes.
    """

    IMG_FRAME_TYPES = {
        dai.Platform.RVC2: dai.ImgFrame.Type.BGR888p,
        dai.Platform.RVC4: dai.ImgFrame.Type.BGR888i,
    }

    def __init__(self) -> None:
        super().__init__()

        self._platform = self.getParentPipeline().getDefaultDevice().getPlatform()

        try:
            self._img_frame_type = self.IMG_FRAME_TYPES[self._platform]
        except KeyError as e:
            raise ValueError(
                f"No dai.ImgFrame.Type defined for platform {self._platform}."
            ) from e

        self._logger = get_logger(self.__class__.__name__)

    def process(self) -> None:
        pass



from typing import List

import cv2
import numpy as np


def nms(dets: np.ndarray, nms_thresh: float = 0.5) -> List[int]:
    """Non-maximum suppression.

    @param dets: Bounding boxes and confidence scores.
    @type dets: np.ndarray
    @param nms_thresh: Non-maximum suppression threshold.
    @type nms_thresh: float
    @return: Indices of the detections to keep.
    @rtype: List[int]
    """
    thresh = nms_thresh
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep


def nms_cv2(
    bboxes: np.ndarray,
    scores: np.ndarray,
    conf_threshold: float,
    iou_threshold: float,
    max_det: int,
):
    """Non-maximum suppression from the opencv-python library.

    @param bboxes: A numpy array of shape (N, 4) containing the bounding boxes.
    @type bboxes: np.ndarray
    @param scores: A numpy array of shape (N,) containing the scores.
    @type scores: np.ndarray
    @param nms_thresh: Non-maximum suppression threshold.
    @type nms_thresh: float
    @return: Indices of the detections to keep.
    @rtype: List[int]
    """

    # NMS
    if len(bboxes) == 0 or len(scores) == 0:
        return []
    keep_indices = cv2.dnn.NMSBoxes(
        bboxes=bboxes.tolist(),
        scores=scores.tolist(),
        score_threshold=conf_threshold,
        nms_threshold=iou_threshold,
        top_k=max_det,
    )

    return keep_indices



import ast
import pickle
from typing import List

import depthai as dai
import pytest

from tests.utils import PipelineMock

from .check_messages import check_output
from .utils import extract_main_slug, extract_parser


@pytest.fixture(scope="session")
def models(request):
    return request.config.getoption("--models")


@pytest.fixture(scope="session")
def duration(request):
    return request.config.getoption("--duration")


@pytest.fixture(scope="session")
def parser_generator():
    from depthai_nodes.node.parser_generator import ParserGenerator

    pipeline = PipelineMock()
    return pipeline.create(ParserGenerator)


def get_parametrized_values(
    models: List[str], parsers: List[str], duration: int
) -> List[List[str]]:
    test_cases = []

    if duration:
        duration = int(duration)

    if models:
        models = ast.literal_eval(models)
        parsers = ast.literal_eval(parsers)
        test_cases.extend(
            [(model, parsers[i], duration) for i, model in enumerate(models)]
        )

    return test_cases


def pytest_generate_tests(metafunc):
    models = metafunc.config.getoption("models")
    parsers = metafunc.config.getoption("parsers")
    duration = metafunc.config.getoption("duration")
    params = get_parametrized_values(models, parsers, duration)
    metafunc.parametrize("model, parser_name, duration", params)


def load_tensors(model: str, parser: str) -> dai.NNData:
    model = extract_main_slug(model)
    nn_data = dai.NNData()
    with open(f"nn_datas/{parser}/{model}.pkl", "rb") as f:
        data = pickle.load(f)
        for key, value in data.items():
            nn_data.addTensor(str(key), value.tolist())
        return nn_data


def test_parser(parser_generator, model: str, parser_name: str, duration: int):
    # Get the model from the HubAI
    try:
        model_description = dai.NNModelDescription(model=model, platform="RVC2")
        archive_path = dai.getModelFromZoo(model_description)
    except Exception:
        try:
            model_description = dai.NNModelDescription(model=model, platform="RVC4")
            archive_path = dai.getModelFromZoo(model_description)
        except Exception as e:
            print(f"Error: {e}")
            exit(7)

    try:
        nn_archive = dai.NNArchive(archive_path)
    except Exception as e:
        print(f"Error: {e}")
        exit(8)

    try:
        parser_name = extract_parser(nn_archive)
    except Exception as e:
        print(e)
        exit(9)

    # Create and build parser
    parser = parser_generator.build(nn_archive=nn_archive)[0]
    parser.input._queue.duration = duration

    # Load and send test data
    try:
        nn_data = load_tensors(model, parser_name)
    except Exception as e:
        print(e)
        exit(10)

    parser.input.send(nn_data)
    parser.out.createOutputQueue(check_output, model, parser_name)  # must be created

    parser.run()



from typing import List

import depthai_nodes.message.creators as creators

from .constants import SCORES


def create_regression(predictions: List[float] = SCORES):
    return creators.create_regression_message(predictions=predictions)



import argparse
import pickle

import depthai as dai
from check_messages import test_output
from utils import extract_main_slug, extract_parser

from depthai_nodes.node import ParserGenerator

argparser = argparse.ArgumentParser()
argparser.add_argument(
    "-m",
    "--model",
    help="The model from which the parser is built",
    required=True,
    type=str,
)
argparser.add_argument("-ip", help="IP address of the device", default="")

args = argparser.parse_args()
model_slug = args.model
IP_mxid = args.ip


def load_tensors(model: str, parser: str) -> dai.NNData:
    model = extract_main_slug(model)
    nn_data = dai.NNData()
    with open(f"nn_datas/{parser}/{model}.pkl", "rb") as f:
        data = pickle.load(f)
        for key, value in data.items():
            nn_data.addTensor(str(key), value.tolist())
        return nn_data


try:
    device = dai.Device(dai.DeviceInfo(IP_mxid))
except Exception as e:
    print(f"Error: {e}")
    exit(6)

device_platform = device.getPlatform().name

# Get the model from the HubAI
try:
    model_description = dai.NNModelDescription(model=model_slug, platform="RVC2")
    archive_path = dai.getModelFromZoo(model_description)
except Exception:
    try:
        model_description = dai.NNModelDescription(model=model_slug, platform="RVC4")
        archive_path = dai.getModelFromZoo(model_description)
    except Exception as e:
        print(f"Error: {e}")
        exit(7)

try:
    nn_archive = dai.NNArchive(archive_path)
except Exception as e:
    print(f"Error: {e}")
    exit(8)

try:
    parser_name = extract_parser(nn_archive)
except Exception as e:
    print(e)
    exit(9)


class Sender(dai.node.ThreadedHostNode):
    def __init__(self):
        dai.node.ThreadedHostNode.__init__(self)
        self.out = self.createOutput()

    def run(self):
        try:
            nnData = load_tensors(model_slug, parser_name)
        except Exception as e:
            nnData = dai.NNData()
            print(e)
            return 10
        self.out.send(nnData)


with dai.Pipeline(device) as pipeline:
    parser = pipeline.create(ParserGenerator).build(nn_archive=nn_archive)[0]

    sender = pipeline.create(Sender)

    sender.out.link(parser.input)
    parser_queue = parser.out.createOutputQueue()

    pipeline.start()

    while pipeline.isRunning():
        message = parser_queue.get()
        test_output(message, model_slug, parser_name)
        pipeline.stop()



import time

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.node import Tiling
from tests.utils import LOG_INTERVAL, OutputMock, create_img_frame


@pytest.fixture
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def img_frame():
    return create_img_frame()


@pytest.fixture
def tiling():
    return Tiling()


def test_build(tiling):
    output = OutputMock()
    grid_size = (2, 2)
    img_shape = (1280, 720)
    nn_shape = np.array([300, 300])
    overlap = 0.1

    tiling.build(
        overlap=overlap,
        img_output=output,
        grid_size=grid_size,
        img_shape=img_shape,
        nn_shape=nn_shape,
        global_detection=False,
    )

    assert tiling.overlap == overlap
    assert tiling.grid_size == grid_size
    np.testing.assert_allclose(tiling.nn_shape, nn_shape)
    assert tiling.img_shape == img_shape

    n_tiles_w, n_tiles_h = grid_size
    A = np.array(
        [
            [n_tiles_w * (1 - overlap) + overlap, 0],
            [0, n_tiles_h * (1 - overlap) + overlap],
        ]
    )
    b = np.array(img_shape)
    expected_tile_dims = np.linalg.inv(A).dot(b)
    np.testing.assert_allclose(tiling.x, expected_tile_dims)

    expected_num_tiles = grid_size[0] * grid_size[1]
    assert len(tiling.tile_positions) == expected_num_tiles


def test_build_global_detection(tiling, img_frame):
    output = OutputMock()
    grid_size = (2, 2)
    img_shape = (img_frame.getWidth(), img_frame.getHeight())
    nn_shape = np.array([300, 300])
    overlap = 0.1

    tiling.build(
        overlap=overlap,
        img_output=output,
        grid_size=grid_size,
        img_shape=img_shape,
        nn_shape=nn_shape,
        global_detection=True,
    )

    first_tile = tiling.tile_positions[0]
    assert first_tile["coords"] == (0, 0, img_shape[0], img_shape[1])

    scale = min(nn_shape[0] / img_shape[0], nn_shape[1] / img_shape[1])
    expected_scaled_width = int(img_shape[0] * scale)
    expected_scaled_height = int(img_shape[1] * scale)
    assert first_tile["scaled_size"] == (expected_scaled_width, expected_scaled_height)


def test_process(tiling: Tiling, img_frame: dai.ImgFrame, duration: float):
    output = OutputMock()
    grid_size = (2, 2)
    img_shape = (img_frame.getWidth(), img_frame.getHeight())
    nn_shape = np.array([300, 300])
    overlap = 0.1
    out_q = tiling.out.createOutputQueue()

    tiling.build(
        overlap=overlap,
        img_output=output,
        grid_size=grid_size,
        img_shape=img_shape,
        nn_shape=nn_shape,
        global_detection=False,
    )

    tiling.process(img_frame)

    expected_num_tiles = len(tiling.tile_positions)

    for _ in range(expected_num_tiles):
        out_frame = out_q.get()
        assert isinstance(out_frame, dai.ImgFrame)

    assert out_q.is_empty()

    if duration:
        start_time = time.time()

        last_log_time = time.time()
        while time.time() - start_time < duration:
            if time.time() - last_log_time > LOG_INTERVAL:
                print(
                    f"Test running... {time.time()-start_time:.1f}s elapsed, {duration-time.time()+start_time:.1f}s remaining"
                )
                last_log_time = time.time()
            tiling.process(img_frame)
            for _ in range(expected_num_tiles):
                out_frame = out_q.get()
                assert isinstance(out_frame, dai.ImgFrame)
            assert out_q.is_empty()



from typing import Any, Dict, List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_segmentation_message
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils.fastsam import (
    box_prompt,
    decode_fastsam_output,
    merge_masks,
    point_prompt,
    process_single_mask,
)
from depthai_nodes.node.parsers.utils.masks_utils import get_segmentation_outputs


class FastSAMParser(BaseParser):
    """Parser class for parsing the output of the FastSAM model.

    Attributes
    ----------
    conf_threshold : float
        Confidence score threshold for detected faces.
    n_classes : int
        Number of classes in the model.
    iou_threshold : float
        Non-maximum suppression threshold.
    mask_conf : float
        Mask confidence threshold.
    prompt : str
        Prompt type.
    points : Tuple[int, int]
        Points.
    point_label : int
        Point label.
    bbox : Tuple[int, int, int, int]
        Bounding box.
    yolo_outputs : List[str]
        Names of the YOLO outputs.
    mask_outputs : List[str]
        Names of the mask outputs.
    protos_output : str
        Name of the protos output.

    Output Message/s
    ----------------
    **Type**: SegmentationMask

    **Description**: SegmentationMask message containing the resulting segmentation masks given the prompt.

    Error Handling
    --------------
    """

    def __init__(
        self,
        conf_threshold: int = 0.5,
        n_classes: int = 1,
        iou_threshold: float = 0.5,
        mask_conf: float = 0.5,
        prompt: str = "everything",
        points: Optional[Tuple[int, int]] = None,
        point_label: Optional[int] = None,
        bbox: Optional[Tuple[int, int, int, int]] = None,
        yolo_outputs: List[str] = None,
        mask_outputs: List[str] = None,
        protos_output: str = "protos_output",
    ) -> None:
        """Initializes the parser node.

        @param conf_threshold: The confidence threshold for the detections
        @type conf_threshold: float
        @param n_classes: The number of classes in the model
        @type n_classes: int
        @param iou_threshold: The intersection over union threshold
        @type iou_threshold: float
        @param mask_conf: The mask confidence threshold
        @type mask_conf: float
        @param prompt: The prompt type
        @type prompt: str
        @param points: The points
        @type points: Optional[Tuple[int, int]]
        @param point_label: The point label
        @type point_label: Optional[int]
        @param bbox: The bounding box
        @type bbox: Optional[Tuple[int, int, int, int]]
        @param yolo_outputs: The YOLO outputs
        @type yolo_outputs: List[str]
        @param mask_outputs: The mask outputs
        @type mask_outputs: List[str]
        @param protos_output: The protos output
        @type protos_output: str
        """
        super().__init__()
        self.conf_threshold = conf_threshold
        self.n_classes = n_classes
        self.iou_threshold = iou_threshold
        self.mask_conf = mask_conf
        self.prompt = prompt
        self.points = points
        self.point_label = point_label
        self.bbox = bbox
        self.yolo_outputs = (
            ["output1_yolov8", "output2_yolov8", "output3_yolov8"]
            if yolo_outputs is None
            else yolo_outputs
        )
        self.mask_outputs = (
            ["output1_masks", "output2_masks", "output3_masks"]
            if mask_outputs is None
            else mask_outputs
        )
        self.protos_output = protos_output
        self._logger.debug(
            f"FastSAMParser initialized with conf_threshold={conf_threshold}, n_classes={n_classes}, iou_threshold={iou_threshold}, mask_conf={mask_conf}, prompt='{prompt}', points={points}, point_label={point_label}, bbox={bbox}, yolo_outputs={yolo_outputs}, mask_outputs={mask_outputs}, protos_output='{protos_output}'"
        )

    def setConfidenceThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold.

        @param threshold: Confidence score threshold.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence threshold must be a float.")
        if threshold < 0 or threshold > 1:
            raise ValueError("Confidence threshold must be between 0 and 1.")
        self.conf_threshold = threshold
        self._logger.debug(f"Confidence threshold set to {self.conf_threshold}")

    def setNumClasses(self, n_classes: int) -> None:
        """Sets the number of classes in the model.

        @param numClasses: The number of classes in the model.
        @type numClasses: int
        """
        if not isinstance(n_classes, int):
            raise ValueError("Number of classes must be an integer.")
        self.n_classes = n_classes
        self._logger.debug(f"Number of classes set to {self.n_classes}")

    def setIouThreshold(self, iou_threshold: float) -> None:
        """Sets the intersection over union threshold.

        @param iou_threshold: The intersection over union threshold.
        @type iou_threshold: float
        """
        if not isinstance(iou_threshold, float):
            raise ValueError("IOU threshold must be a float.")
        if iou_threshold < 0 or iou_threshold > 1:
            raise ValueError("IOU threshold must be between 0 and 1.")
        self.iou_threshold = iou_threshold
        self._logger.debug(
            f"Intersection over union threshold set to {self.iou_threshold}"
        )

    def setMaskConfidence(self, mask_conf: float) -> None:
        """Sets the mask confidence threshold.

        @param mask_conf: The mask confidence threshold.
        @type mask_conf: float
        """
        if not isinstance(mask_conf, float):
            raise ValueError("Mask confidence must be a float.")
        if mask_conf < 0 or mask_conf > 1:
            raise ValueError("Mask confidence must be between 0 and 1.")
        self.mask_conf = mask_conf
        self._logger.debug(f"Mask confidence threshold set to {self.mask_conf}")

    def setPrompt(self, prompt: str) -> None:
        """Sets the prompt type.

        @param prompt: The prompt type
        @type prompt: str
        """
        if not isinstance(prompt, str):
            raise ValueError("Prompt must be a string.")
        if prompt not in ["everything", "bbox", "point"]:
            raise ValueError("Prompt must be one of 'everything', 'bbox', or 'point'")
        self.prompt = prompt
        self._logger.debug(f"Prompt set to '{self.prompt}'")

    def setPoints(self, points: Tuple[int, int]) -> None:
        """Sets the points.

        @param points: The points
        @type points: Tuple[int, int]
        """
        if not isinstance(points, tuple):
            raise ValueError("Points must be a tuple.")
        if len(points) != 2:
            raise ValueError("Points must be a tuple of length 2.")
        if not all(isinstance(p, int) for p in points):
            raise ValueError("Point elements must be integers.")
        self.points = points
        self._logger.debug(f"Points set to {self.points}")

    def setPointLabel(self, point_label: int) -> None:
        """Sets the point label.

        @param point_label: The point label
        @type point_label: int
        """
        if not isinstance(point_label, int):
            raise ValueError("Point label must be an integer.")
        self.point_label = point_label
        self._logger.debug(f"Point label set to {self.point_label}")

    def setBoundingBox(self, bbox: Tuple[int, int, int, int]) -> None:
        """Sets the bounding box.

        @param bbox: The bounding box
        @type bbox: Tuple[int, int, int, int]
        """
        if not isinstance(bbox, tuple):
            raise ValueError("Bounding box must be a tuple.")
        if len(bbox) != 4:
            raise ValueError("Bounding box must be a tuple of length 4.")
        if not all(isinstance(b, int) for b in bbox):
            raise ValueError("Bounding box elements must be integers.")
        self.bbox = bbox
        self._logger.debug(f"Bounding box set to {self.bbox}")

    def setYoloOutputs(self, yolo_outputs: List[str]) -> None:
        """Sets the YOLO outputs.

        @param yolo_outputs: The YOLO outputs
        @type yolo_outputs: List[str]
        """
        if not isinstance(yolo_outputs, list):
            raise ValueError("YOLO outputs must be a list.")
        if not all(isinstance(o, str) for o in yolo_outputs):
            raise ValueError("YOLO outputs must be a list of strings.")
        self.yolo_outputs = yolo_outputs
        self._logger.debug(f"YOLO outputs set to {self.yolo_outputs}")

    def setMaskOutputs(self, mask_outputs: List[str]) -> None:
        """Sets the mask outputs.

        @param mask_outputs: The mask outputs
        @type mask_outputs: List[str]
        """
        if not isinstance(mask_outputs, list):
            raise ValueError("Mask outputs must be a list.")
        if not all(isinstance(o, str) for o in mask_outputs):
            raise ValueError("Mask outputs must be a list of strings.")
        self.mask_outputs = mask_outputs
        self._logger.debug(f"Mask outputs set to {self.mask_outputs}")

    def setProtosOutput(self, protos_output: str) -> None:
        """Sets the protos output.

        @param protos_output: The protos output
        @type protos_output: str
        """
        if not isinstance(protos_output, str):
            raise ValueError("Protos output must be a string.")
        self.protos_output = protos_output
        self._logger.debug(f"Protos output set to '{self.protos_output}'")

    def build(self, head_config: Dict[str, Any]) -> "FastSAMParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: FastSAMParser
        """

        output_layers = head_config["outputs"]
        yolo_outputs = [name for name in output_layers if "_yolo" in name]
        if len(yolo_outputs) != 0:
            self.yolo_outputs = yolo_outputs

        mask_outputs = [name for name in output_layers if "_masks" in name]
        if len(mask_outputs) != 0:
            self.mask_outputs = mask_outputs

        self.protos_output = head_config.get("protos_output", self.protos_output)

        self.conf_threshold = head_config.get("conf_threshold", self.conf_threshold)
        self.n_classes = head_config.get("n_classes", self.n_classes)
        self.iou_threshold = head_config.get("iou_threshold", self.iou_threshold)
        self.mask_conf = head_config.get("mask_conf", self.mask_conf)
        self.prompt = head_config.get("prompt", self.prompt)
        self.points = head_config.get("points", self.points)
        self.point_label = head_config.get("point_label", self.point_label)
        self.bbox = head_config.get("bbox", self.bbox)

        self._logger.debug(
            f"FastSAMParser built with conf_threshold={self.conf_threshold}, n_classes={self.n_classes}, iou_threshold={self.iou_threshold}, mask_conf={self.mask_conf}, prompt='{self.prompt}', points={self.points}, point_label={self.point_label}, bbox={self.bbox}, yolo_outputs={self.yolo_outputs}, mask_outputs={self.mask_outputs}, protos_output='{self.protos_output}'"
        )

        return self

    def run(self):
        self._logger.debug("FastSAMParser run started")
        if self.prompt not in ["everything", "bbox", "point"]:
            raise ValueError("Prompt must be one of 'everything', 'bbox', or 'point'")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data

            outputs_names = sorted([name for name in self.yolo_outputs])
            self._logger.debug(f"Processing input with layers: {outputs_names}")
            outputs_values = [
                output.getTensor(
                    o, dequantize=True, storageOrder=dai.TensorInfo.StorageOrder.NCHW
                ).astype(np.float32)
                for o in outputs_names
            ]
            # Get the segmentation outputs
            (
                masks_outputs_values,
                protos_output,
                protos_len,
            ) = get_segmentation_outputs(output, self.mask_outputs, self.protos_output)

            # determine the input shape of the model from the first output
            width = outputs_values[0].shape[3] * 8
            height = outputs_values[0].shape[2] * 8
            input_shape = (width, height)

            # Decode the outputs
            results = decode_fastsam_output(
                outputs_values,
                [8, 16, 32],
                [None, None, None],
                img_shape=input_shape[::-1],
                conf_thres=self.conf_threshold,
                iou_thres=self.iou_threshold,
                num_classes=self.n_classes,
            )

            bboxes, masks = [], []
            for i in range(results.shape[0]):
                bbox, conf, label, seg_coeff = (
                    results[i, :4].astype(int),
                    results[i, 4],
                    results[i, 5].astype(int),
                    results[i, 6:].astype(int),
                )
                bboxes.append(bbox.tolist() + [conf, int(label)])
                hi, ai, xi, yi = seg_coeff
                mask_coeff = masks_outputs_values[hi][
                    0, ai * protos_len : (ai + 1) * protos_len, yi, xi
                ]
                mask = process_single_mask(
                    protos_output[0], mask_coeff, self.mask_conf, input_shape, bbox
                )
                masks.append(mask)

            results_bboxes = np.array(bboxes)
            results_masks = np.array(masks)

            if self.prompt == "bbox":
                results_masks = box_prompt(
                    results_masks, bbox=self.bbox, orig_shape=input_shape[::-1]
                )
            elif self.prompt == "point":
                results_masks = point_prompt(
                    results_bboxes,
                    results_masks,
                    points=self.points,
                    pointlabel=self.point_label,
                    orig_shape=input_shape[::-1],
                )

            if len(results_masks) == 0:
                results_masks = np.full((1, height, width), -1, dtype=np.int16)
            results_masks = merge_masks(results_masks)

            segmentation_message = create_segmentation_message(results_masks)
            segmentation_message.setTransformation(output.getTransformation())
            segmentation_message.setTimestamp(output.getTimestamp())
            segmentation_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created segmentation message with {len(results_masks)} masks"
            )

            self.out.send(segmentation_message)

            self._logger.debug("Segmentation message sent successfully")



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.message.creators import (
    create_image_message,
)

IMAGE = np.random.randint(0, 256, (480, 640, 3), dtype=np.uint8)
IMAGE_GRAY = np.random.randint(0, 256, (480, 640, 1), dtype=np.uint8)


def test_valid_hwc_bgr():
    img_frame = create_image_message(IMAGE, is_bgr=True)

    assert isinstance(img_frame, dai.ImgFrame)
    assert img_frame.getWidth() == 640
    assert img_frame.getHeight() == 480
    assert img_frame.getType() == dai.ImgFrame.Type.BGR888i
    assert np.array_equal(img_frame.getCvFrame(), IMAGE)


def test_valid_hwc_rgb():
    create_image_message(IMAGE, is_bgr=False)


def test_valid_chw_bgr():
    image = IMAGE.transpose(2, 0, 1)
    create_image_message(image, is_bgr=True)


def test_valid_chw_rgb():
    image = IMAGE.transpose(2, 0, 1)
    create_image_message(image, is_bgr=False)


def test_valid_hwc_grayscale():
    img_frame = create_image_message(IMAGE_GRAY, is_bgr=True)

    assert isinstance(img_frame, dai.ImgFrame)
    assert img_frame.getWidth() == 640
    assert img_frame.getHeight() == 480
    assert img_frame.getType() == dai.ImgFrame.Type.GRAY8


def test_valid_chw_grayscale():
    image = IMAGE_GRAY.transpose(2, 0, 1)
    create_image_message(image, is_bgr=True)


def test_invalid_shape():
    image = np.random.randint(0, 256, (480, 640, 4), dtype=np.uint8)
    with pytest.raises(ValueError):
        create_image_message(image, is_bgr=True)


def test_invalid_dtype():
    image = IMAGE.astype(np.float32)
    with pytest.raises(ValueError):
        create_image_message(image, is_bgr=True)



import cv2
import depthai as dai
import numpy as np


def create_image_message(
    image: np.ndarray,
    is_bgr: bool = True,
    img_frame_type: dai.ImgFrame.Type = dai.ImgFrame.Type.BGR888i,
) -> dai.ImgFrame:
    """Create a DepthAI message for an image array.

    @param image: Image array in HWC or CHW format.
    @type image: np.array
    @param is_bgr: If True, the image is in BGR format. If False, the image is in RGB
        format. Defaults to True.
    @type is_bgr: bool
    @param img_frame_type: Output ImgFrame type. Defaults to BGR888i.
    @type img_frame_type: dai.ImgFrame.Type
    @return: dai.ImgFrame object containing the image information.
    @rtype: dai.ImgFrame
    @raise ValueError: If the image shape is not CHW or HWC.
    """

    if image.shape[0] in [1, 3]:
        hwc = False
    elif image.shape[2] in [1, 3]:
        hwc = True
    else:
        raise ValueError(
            f"Unexpected image shape. Expected CHW or HWC, got {image.shape}"
        )

    if not hwc:
        image = np.transpose(image, (1, 2, 0))

    if isinstance(image[0, 0, 0], (float, np.floating)):
        raise ValueError(f"Expected int type, got {type(image[0, 0, 0])}.")

    if image.shape[2] == 1:  # grayscale
        image = image[:, :, 0]  # HW image
        if not (img_frame_type.name.startswith(("RAW", "GRAY"))):
            img_frame_type = dai.ImgFrame.Type.GRAY8
        height, width = image.shape
    else:
        if not is_bgr:
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        height, width, _ = image.shape

    imgFrame = dai.ImgFrame()
    imgFrame.setCvFrame(image, img_frame_type)
    imgFrame.setWidth(width)
    imgFrame.setHeight(height)
    imgFrame.setType(img_frame_type)

    return imgFrame



import pytest

from depthai_nodes import Keypoint, Keypoints
from depthai_nodes.message.creators import create_keypoints_message

KPTS = [[0.1, 0.2], [0.3, 0.4]]
SCORES = [0.9, 0.8]


def test_valid_2d_keypoints():
    message = create_keypoints_message(KPTS, SCORES)

    assert isinstance(message, Keypoints)
    assert len(message.keypoints) == 2
    assert all(isinstance(kp, Keypoint) for kp in message.keypoints)
    for i, kp in enumerate(message.keypoints):
        assert kp.x == KPTS[i][0]
        assert kp.y == KPTS[i][1]
        assert kp.z == 0.0
        assert kp.confidence == SCORES[i]


def test_valid_3d_keypoints():
    keypoints = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]
    scores = [0.9, 0.8]
    message = create_keypoints_message(keypoints, scores)

    assert isinstance(message, Keypoints)
    assert len(message.keypoints) == 2
    assert all(isinstance(kp, Keypoint) for kp in message.keypoints)
    for i, kp in enumerate(message.keypoints):
        assert kp.x == keypoints[i][0]
        assert kp.y == keypoints[i][1]
        assert kp.z == keypoints[i][2]
        assert kp.confidence == scores[i]


def test_valid_keypoints_no_scores():
    message = create_keypoints_message(KPTS)

    assert isinstance(message, Keypoints)
    assert len(message.keypoints) == 2
    assert all(isinstance(kp, Keypoint) for kp in message.keypoints)
    for kp in message.keypoints:
        assert kp.confidence == -1


def test_invalid_keypoints_type():
    with pytest.raises(ValueError):
        create_keypoints_message("not a list or array")


def test_invalid_scores_type():
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, scores="not a list or array")


def test_mismatched_keypoints_scores_length():
    scores = [0.9]
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, scores)


def test_invalid_scores_values():
    scores = [0.9, "not a float"]
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, scores)


def test_scores_out_of_range():
    scores = [0.9, 1.1]
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, scores)


def test_invalid_confidence_threshold_type():
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, SCORES, confidence_threshold="not a float")


def test_confidence_threshold_out_of_range():
    with pytest.raises(ValueError):
        create_keypoints_message(KPTS, SCORES, confidence_threshold=1.1)


def test_invalid_keypoints_shape():
    keypoints = [[0.1, 0.2, 0.3, 0.4]]
    with pytest.raises(ValueError):
        create_keypoints_message(keypoints)


def test_invalid_keypoints_inner_type():
    keypoints = [[0.1, "not a float"]]
    with pytest.raises(ValueError):
        create_keypoints_message(keypoints)


def test_all_keypoints_same_shape():
    keypoints = [[0.1, 0.2], [0.3, 0.4, 0.5]]
    with pytest.raises(ValueError):
        create_keypoints_message(keypoints)



from typing import Any, Dict, List

import cv2
import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_detection_message
from depthai_nodes.node.parsers.detection import DetectionParser
from depthai_nodes.node.parsers.utils.medipipe import generate_anchors_and_decode


class MPPalmDetectionParser(DetectionParser):
    """Parser class for parsing the output of the Mediapipe Palm detection model. As the
    result, the node sends out the detected hands in the form of a message containing
    bounding boxes, labels, and confidence scores.

    Attributes
    ----------
    output_layer_names: List[str]
        Names of the output layers relevant to the parser.
    conf_threshold : float
        Confidence score threshold for detected hands.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.
    scale : int
        Scale of the input image.

    Output Message/s
    -------
    **Type**: ImgDetectionsExtended

    **Description**: ImgDetectionsExtended message containing bounding boxes, labels, and confidence scores of detected hands.

    See also
    --------
    Official MediaPipe Hands solution:
    https://ai.google.dev/edge/mediapipe/solutions/vision/hand_landmarker
    """

    def __init__(
        self,
        output_layer_names: List[str] = None,
        conf_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        max_det: int = 100,
        scale: int = 192,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_names: Names of the output layers relevant to the parser.
        @type output_layer_names: List[str]
        @param conf_threshold: Confidence score threshold for detected hands.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        @param scale: Scale of the input image.
        @type scale: int
        """
        super().__init__(conf_threshold, iou_threshold, max_det)
        self.output_layer_names = (
            [] if output_layer_names is None else output_layer_names
        )
        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det
        self.scale = scale
        self.label_names = ["Palm"]
        self._logger.debug(
            f"MPPalmDetectionParser initialized with output_layer_names={output_layer_names}, conf_threshold={conf_threshold}, iou_threshold={iou_threshold}, max_det={max_det}, scale={scale}"
        )

    def setOutputLayerNames(self, output_layer_names: List[str]) -> None:
        """Sets the output layer name(s) for the parser.

        @param output_layer_names: The name of the output layer(s) from which the scores
            are extracted.
        @type output_layer_names: List[str]
        """
        if not isinstance(output_layer_names, list):
            raise ValueError("Output layer name must be a list.")
        if not all(isinstance(layer_name, str) for layer_name in output_layer_names):
            raise ValueError("Each output layer name must be a string.")
        if len(output_layer_names) != 2:
            raise ValueError(
                f"Only two output layers are supported for MPPalmDetectionParser, got {len(output_layer_names)} layers."
            )
        self.output_layer_names = output_layer_names
        self._logger.debug(f"Output layer names set to {self.output_layer_names}")

    def setScale(self, scale: int) -> None:
        """Sets the scale of the input image.

        @param scale: Scale of the input image.
        @type scale: int
        """
        if not isinstance(scale, int):
            raise ValueError("Scale must be an integer.")
        self.scale = scale
        self._logger.debug(f"Scale set to {self.scale}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "MPPalmDetectionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: MPPalmDetectionParser
        """

        super().build(head_config)
        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 2:
            raise ValueError(
                f"Only two output layers are supported for MPPalmDetectionParser, got {len(output_layers)} layers."
            )
        self.output_layer_names = output_layers
        self.scale = head_config.get("scale", self.scale)

        self._logger.debug(
            f"MPPalmDetectionParser built with output_layer_names={self.output_layer_names}, scale={self.scale}"
        )

        return self

    def run(self):
        self._logger.debug("MPPalmDetectionParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            all_tensors = output.getAllLayerNames()

            self._logger.debug(f"Processing input with layers: {all_tensors}")

            bboxes = None
            scores = None

            for tensor_name in all_tensors:
                tensor = np.array(
                    output.getTensor(tensor_name, dequantize=True), dtype=np.float32
                )

                if bboxes is None:
                    bboxes = tensor
                    scores = tensor
                else:
                    bboxes = bboxes if tensor.shape[-1] < bboxes.shape[-1] else tensor
                    scores = tensor if tensor.shape[-1] < scores.shape[-1] else scores

            bboxes = bboxes.reshape(-1, 18)
            scores = scores.reshape(-1)

            if bboxes is None or scores is None:
                raise ValueError("No valid output tensors found.")

            decoded_bboxes = generate_anchors_and_decode(
                bboxes=bboxes,
                scores=scores,
                threshold=self.conf_threshold,
                scale=self.scale,
            )

            bboxes = []
            scores = []
            angles = []
            for hand in decoded_bboxes:
                extended_points = np.array(hand.rect_points)

                x_dist = extended_points[3][0] - extended_points[0][0]
                y_dist = extended_points[3][1] - extended_points[0][1]

                angle = np.degrees(np.arctan2(y_dist, x_dist))
                x_center, y_center = np.mean(extended_points, axis=0)
                width = np.linalg.norm(extended_points[0] - extended_points[3])
                height = np.linalg.norm(extended_points[0] - extended_points[1])

                bboxes.append([x_center, y_center, width, height])
                angles.append(angle)
                scores.append(hand.pd_score)

            indices = cv2.dnn.NMSBoxes(
                bboxes,
                scores,
                self.conf_threshold,
                self.iou_threshold,
                top_k=self.max_det,
            )
            bboxes = np.array(bboxes)[indices]
            scores = np.array(scores)[indices]
            angles = np.array(angles)[indices]
            bboxes = bboxes.astype(float) / self.scale

            bboxes = np.clip(bboxes, 0, 1)
            angles = np.round(angles, 0)

            labels = np.array([0] * len(bboxes))

            label_names = (
                [self.label_names[label] for label in labels]
                if self.label_names
                else None
            )
            detections_msg = create_detection_message(
                bboxes=bboxes,
                scores=scores,
                angles=angles,
                labels=labels,
                label_names=label_names,
            )
            detections_msg.setTimestamp(output.getTimestamp())
            detections_msg.setTransformation(output.getTransformation())
            detections_msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created detection message with {len(bboxes)} detections"
            )

            self.out.send(detections_msg)

            self._logger.debug("Detection message sent successfully")



import numpy as np
import pytest

from depthai_nodes import Classifications
from depthai_nodes.message.creators import (
    create_classification_message,
)

CLASSES = ["cat", "dog", "bird"]
SCORES = [0.7, 0.2, 0.1]


def test_valid_input():
    message = create_classification_message(CLASSES, SCORES)

    assert isinstance(message, Classifications)
    assert message.classes == ["cat", "dog", "bird"]
    assert np.array_equal(message.scores, np.array(SCORES, dtype=np.float32))


def test_single_class_and_score():
    classes = ["cat"]
    scores = [1.0]

    message = create_classification_message(classes, scores)
    assert message.classes == ["cat"]
    assert message.scores == [1.0]


def test_identical_scores():
    scores = [1 / 3, 1 / 3, 1 / 3]

    message = create_classification_message(CLASSES, scores)

    assert message.classes == CLASSES
    assert np.all(message.scores == np.array(scores, dtype=np.float32))


def test_duplicate_scores():
    scores = [0.4, 0.2, 0.4]

    correct_classes = ["cat", "bird", "dog"]
    correct_scores = [0.4, 0.4, 0.2]

    message = create_classification_message(CLASSES, scores)

    assert message.classes == correct_classes
    assert np.all(message.scores == np.array(correct_scores, dtype=np.float32))


def test_very_small_scores():
    scores = [1e-10, 1e-10, 1 - 2e-10]

    message = create_classification_message(CLASSES, scores)

    assert isinstance(message, Classifications)
    assert message.classes == ["bird", "cat", "dog"]
    assert np.all(
        message.scores == np.array([1 - 2e-10, 1e-10, 1e-10], dtype=np.float32)
    )


def test_none_classes():
    with pytest.raises(ValueError):
        create_classification_message(None, SCORES)


def test_non_list_classes():
    with pytest.raises(ValueError):
        create_classification_message("not a list", SCORES)


def test_empty_classes():
    with pytest.raises(ValueError):
        create_classification_message([], SCORES)


def test_none_scores():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, None)


def test_non_list_non_array_scores():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, "not a list or array")


def test_empty_scores():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, [])


def test_non_1d_array_scores():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, np.array([SCORES]))


def test_non_float_scores():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, [0.7, 0.2, "not a float"])


def test_scores_not_sum_to_1():
    with pytest.raises(ValueError):
        create_classification_message(CLASSES, [0.2, 0.2, 0.2])


def test_mismatched_classes_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog"], [0.7, 0.2, 0.1])



from typing import List, Tuple

import numpy as np


def decode_scores_and_points(
    tpMap: np.ndarray, heat: np.ndarray, topk_n: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Decode the scores and points from the neural network output tensors. Used for
    MLSD model.

    @param tpMap: Tensor containing the vector map.
    @type tpMap: np.ndarray
    @param heat: Tensor containing the heat map.
    @type heat: np.ndarray
    @param topk_n: Number of top candidates to keep.
    @type topk_n: int
    @return: Detected points, confidence scores for the detected points, and vector map.
    @rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    b, c, h, w = tpMap.shape
    displacement = tpMap[:, 1:5, :, :][0]

    indices_np = np.argpartition(heat, -topk_n)[-topk_n:]
    pts_score = heat[indices_np]
    yy_np = np.floor_divide(indices_np, w).reshape(-1, 1)
    xx_np = np.fmod(indices_np, w).reshape(-1, 1)
    pts = np.hstack((yy_np, xx_np))

    vmap = displacement.transpose((1, 2, 0))
    return pts, pts_score, vmap


def get_lines(
    pts: np.ndarray,
    pts_score: np.ndarray,
    vmap: np.ndarray,
    score_thr: float,
    dist_thr: float,
    input_size: int = 512,
) -> Tuple[np.ndarray, List[float]]:
    """Get lines from the detected points and scores. The lines are filtered by the
    score threshold and distance threshold. Used for MLSD model.

    @param pts: Detected points.
    @type pts: np.ndarray
    @param pts_score: Confidence scores for the detected points.
    @type pts_score: np.ndarray
    @param vmap: Vector map.
    @type vmap: np.ndarray
    @param score_thr: Confidence score threshold for detected lines.
    @type score_thr: float
    @param dist_thr: Distance threshold for merging lines.
    @type dist_thr: float
    @param input_size: Input size of the model.
    @type input_size: int
    @return: Detected lines and their confidence scores.
    @rtype: Tuple[np.ndarray, List[float]]
    """
    start = vmap[:, :, :2]
    end = vmap[:, :, 2:]
    dist_map = np.sqrt(np.sum((start - end) ** 2, axis=-1))

    segments_list = []
    line_scores = []
    for center, score in zip(pts, pts_score):
        y, x = center
        distance = dist_map[y, x]
        if score > score_thr and distance > dist_thr:
            disp_x_start, disp_y_start, disp_x_end, disp_y_end = vmap[y, x, :]
            x_start = x + disp_x_start
            y_start = y + disp_y_start
            x_end = x + disp_x_end
            y_end = y + disp_y_end
            segments_list.append([x_start, y_start, x_end, y_end])
            line_scores.append(score)

    lines = 2 * np.array(segments_list)  # 256 > 512
    lines /= input_size
    return lines, line_scores



from typing import Any, Dict, List

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import (
    create_classification_message,
)
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils import softmax


class ClassificationParser(BaseParser):
    """Postprocessing logic for Classification model.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    classes : List[str]
        List of class names to be used for linking with their respective scores.
        Expected to be in the same order as Neural Network's output. If not provided, the message will only return sorted scores.
    is_softmax : bool = True
        If False, the scores are converted to probabilities using softmax function.

    Output Message/s
    ----------------
    **Type** : Classifications(dai.Buffer)

    **Description**: An object with attributes `classes` and `scores`. `classes` is a list of classes, sorted in descending order of scores. `scores` is a list of corresponding scores.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        classes: List[str] = None,
        is_softmax: bool = True,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param classes: List of class names to be used for linking with their respective
            scores. Expected to be in the same order as Neural Network's output. If not
            provided, the message will only return sorted scores.
        @type classes: List[str]
        @param is_softmax: If False, the scores are converted to probabilities using
            softmax function.
        @type is_softmax: bool
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self.classes = classes or []
        self.n_classes = len(self.classes)
        self.is_softmax = is_softmax
        self._logger.debug(
            f"ClassificationParser initialized with output_layer_name='{output_layer_name}', classes={classes}, is_softmax={is_softmax}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setClasses(self, classes: List[str]) -> None:
        """Sets the class names for the classification model.

        @param classes: List of class names to be used for linking with their respective
            scores.
        @type classes: List[str]
        """
        if not isinstance(classes, list):
            raise ValueError("classes must be a list.")
        for class_name in classes:
            if not isinstance(class_name, str):
                raise ValueError("Each class name must be a string.")
        self.classes = classes if classes is not None else []
        self.n_classes = len(self.classes)
        self._logger.debug(f"Classes set to {self.classes}")

    def setSoftmax(self, is_softmax: bool) -> None:
        """Sets the softmax flag for the classification model.

        @param is_softmax: If False, the parser will convert the scores to probabilities
            using softmax function.
        @type is_softmax: bool
        """
        if not isinstance(is_softmax, bool):
            raise ValueError("is_softmax must be a boolean.")
        self.is_softmax = is_softmax
        self._logger.debug(f"Softmax set to {self.is_softmax}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "ClassificationParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: ClassificationParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for Classification, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.classes = head_config.get("classes", self.classes)
        self.n_classes = head_config.get("n_classes", self.n_classes)
        self.is_softmax = head_config.get("is_softmax", self.is_softmax)

        self._logger.debug(
            f"ClassificationParser built with output_layer_name='{self.output_layer_name}', classes={self.classes}, n_classes={self.n_classes}, is_softmax={self.is_softmax}"
        )
        return self

    def run(self):
        self._logger.debug("ClassificationParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            scores = output.getTensor(self.output_layer_name, dequantize=True)
            scores = np.array(scores).flatten()

            if len(scores) != self.n_classes and self.n_classes != 0:
                raise ValueError(
                    f"Number of labels and scores mismatch. Provided {self.n_classes} class names and {len(scores)} scores."
                )

            if not self.is_softmax:
                scores = softmax(scores)

            msg = create_classification_message(self.classes, scores)
            msg.setTransformation(output.getTransformation())
            msg.setTimestamp(output.getTimestamp())
            msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(f"Created message with {len(msg.classes)} classes")

            self.out.send(msg)

            self._logger.debug("Classification message sent successfully")



from typing import List, Optional, Tuple, Union

import numpy as np

from depthai_nodes import Keypoint, Keypoints


def create_keypoints_message(
    keypoints: Union[np.ndarray, List[List[float]]],
    scores: Union[np.ndarray, List[float], None] = None,
    confidence_threshold: Optional[float] = None,
    label_names: Optional[List[str]] = None,
    edges: Optional[List[Tuple[int, int]]] = None,
) -> Keypoints:
    """Create a DepthAI message for the keypoints.

    NOTE: If you provide a confidence threshold for filtering the keypoints based on scores, the edges will be filtered to only include the edges between the keypoints that are present in the filtered keypoints. This is done to ensure that the edges are only drawn between the keypoints that are present in the filtered keypoints. You can always access the full set of edges in the model's NN archive.

    @param keypoints: Detected 2D or 3D keypoints of shape (N,2 or 3) meaning [...,[x, y],...] or [...,[x, y, z],...].
    @type keypoints: np.ndarray or List[List[float]]
    @param scores: Confidence scores of the detected keypoints. Defaults to None.
    @type scores: Union[np.ndarray, List[float], None]
    @param confidence_threshold: Confidence threshold of keypoint detections. Defaults to None.
    @type confidence_threshold: Optional[float]
    @param label_names: label_names of the detected keypoints. Defaults to None.
    @type label_names: Optional[List[str]]
    @param edges: Connection pairs of the detected keypoints. Defaults to None. Example: [[0,1], [1,2], [2,3], [3,0]] shows that keypoint 0 is connected to keypoint 1, keypoint 1 is connected to keypoint 2, etc.
    @type edges: Optional[List[Tuple[int, int]]]
    @return: Keypoints message containing the detected keypoints.
    @rtype: Keypoints

    @raise ValueError: If the keypoints are not a numpy array or list.
    @raise ValueError: If the scores are not a numpy array or list.
    @raise ValueError: If scores and keypoints do not have the same length.
    @raise ValueError: If score values are not floats.
    @raise ValueError: If score values are not between 0 and 1.
    @raise ValueError: If the confidence threshold is not a float.
    @raise ValueError: If the confidence threshold is not between 0 and 1.
    @raise ValueError: If the keypoints are not of shape (N,2 or 3).
    @raise ValueError: If the keypoints 2nd dimension is not of size E{2} or E{3}.
    """

    if not isinstance(keypoints, (np.ndarray, list)):
        raise ValueError(
            f"Keypoints should be numpy array or list, got {type(keypoints)}."
        )

    if scores is not None:
        if not isinstance(scores, (np.ndarray, list)):
            raise ValueError(
                f"Scores should be numpy array or list, got {type(scores)}."
            )

        if len(keypoints) != len(scores):
            raise ValueError(
                "Keypoints and scores should have the same length. Got {} keypoints and {} scores.".format(
                    len(keypoints), len(scores)
                )
            )

        if not all(isinstance(score, (float, np.floating)) for score in scores):
            raise ValueError("Scores should only contain float values.")
        if not all(0 <= score <= 1 for score in scores):
            raise ValueError("Scores should only contain values between 0 and 1.")

    if confidence_threshold is not None:
        if not isinstance(confidence_threshold, float):
            raise ValueError(
                f"The confidence_threshold should be float, got {type(confidence_threshold)}."
            )

        if not (0 <= confidence_threshold <= 1):
            raise ValueError(
                f"The confidence_threshold should be between 0 and 1, got confidence_threshold {confidence_threshold}."
            )

    if len(keypoints) != 0:
        dimension = len(keypoints[0])
        if dimension != 2 and dimension != 3:
            raise ValueError(
                f"All keypoints should be of dimension 2 or 3, got dimension {dimension}."
            )

    if isinstance(keypoints, list):
        for keypoint in keypoints:
            if not isinstance(keypoint, list):
                raise ValueError(
                    f"Keypoints should be list of lists or np.array, got list of {type(keypoint)}."
                )
            if len(keypoint) != dimension:
                raise ValueError(
                    "All keypoints have to be of same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
                )
            for coord in keypoint:
                if not isinstance(coord, (float, np.floating)):
                    raise ValueError(
                        f"Keypoints inner list should contain only float, got {type(coord)}."
                    )

    if label_names is not None:
        if not isinstance(label_names, list):
            raise ValueError(f"label_names should be list, got {type(label_names)}.")
        if not all(isinstance(label, str) for label in label_names):
            raise ValueError("label_names should be a list of strings.")

    if edges is not None:
        if not isinstance(edges, list):
            raise ValueError(f"Edges should be list, got {type(edges)}.")
        if not all(
            isinstance(edge, tuple)
            and len(edge) == 2
            and all(isinstance(i, int) for i in edge)
            for edge in edges
        ):
            raise ValueError("Edges should be a list of tuples of integers.")

    keypoints = np.array(keypoints)
    if scores is not None:
        scores = np.array(scores)

    if len(keypoints) != 0:
        if len(keypoints.shape) != 2:
            raise ValueError(
                f"Keypoints should be of shape (N,2 or 3) got {keypoints.shape}."
            )

        use_3d = keypoints.shape[1] == 3

    keypoints_msg = Keypoints()
    points = []
    included_keypoints = []

    for i, keypoint in enumerate(keypoints):
        if scores is not None and confidence_threshold is not None:
            if scores[i] < confidence_threshold:
                continue
        pt = Keypoint()
        pt.x = float(keypoint[0])
        pt.y = float(keypoint[1])
        pt.z = float(keypoint[2]) if use_3d else 0.0
        if scores is not None:
            pt.confidence = float(scores[i])
        if label_names is not None:
            pt.label_name = label_names[i]
        points.append(pt)
        included_keypoints.append(i)

    keypoints_msg.keypoints = points

    if edges is not None:
        filtered_edges = []
        # Create a mapping from original indices to new indices
        index_mapping = {
            old_idx: new_idx for new_idx, old_idx in enumerate(included_keypoints)
        }

        for edge in edges:
            # Only include edges where both endpoints exist in the filtered keypoints
            if edge[0] in included_keypoints and edge[1] in included_keypoints:
                # Map the old indices to the new indices
                new_edge = [index_mapping[edge[0]], index_mapping[edge[1]]]
                filtered_edges.append(new_edge)

        keypoints_msg.edges = filtered_edges
    return keypoints_msg



from typing import List

import numpy as np

import depthai_nodes.message.creators as creators

from .constants import COLLECTIONS, SCORES


def create_lines(
    lines: np.ndarray = COLLECTIONS["lines"],
    scores: List[float] = SCORES,
):
    return creators.create_line_detection_message(
        lines=np.array(lines), scores=np.array(scores)
    )



import depthai as dai
import numpy as np


def create_feature_point(x: float, y: float, id: int, age: int) -> dai.TrackedFeature:
    """Create a tracked feature point.

    @param x: X coordinate of the feature point.
    @type x: float
    @param y: Y coordinate of the feature point.
    @type y: float
    @param id: ID of the feature point.
    @type id: int
    @param age: Age of the feature point.
    @type age: int
    @return: Tracked feature point.
    @rtype: dai.TrackedFeature
    """

    feature = dai.TrackedFeature()
    feature.position.x = x
    feature.position.y = y
    feature.id = id
    feature.age = age

    return feature


def create_tracked_features_message(
    reference_points: np.ndarray, target_points: np.ndarray
) -> dai.TrackedFeatures:
    """Create a DepthAI message for tracked features.

    @param reference_points: Reference points of shape (N,2) meaning [...,[x, y],...].
    @type reference_points: np.ndarray
    @param target_points: Target points of shape (N,2) meaning [...,[x, y],...].
    @type target_points: np.ndarray

    @return: Message containing the tracked features.
    @rtype: dai.TrackedFeatures

    @raise ValueError: If the reference_points are not a numpy array.
    @raise ValueError: If the reference_points are not of shape (N,2).
    @raise ValueError: If the reference_points 2nd dimension is not of size E{2}.
    @raise ValueError: If the target_points are not a numpy array.
    @raise ValueError: If the target_points are not of shape (N,2).
    @raise ValueError: If the target_points 2nd dimension is not of size E{2}.
    """

    if not isinstance(reference_points, np.ndarray):
        raise ValueError(
            f"reference_points should be numpy array, got {type(reference_points)}."
        )
    if len(reference_points.shape) != 2:
        raise ValueError(
            f"reference_points should be of shape (N,2) meaning [...,[x, y],...], got {reference_points.shape}."
        )
    if reference_points.shape[1] != 2:
        raise ValueError(
            f"reference_points 2nd dimension should be of size 2 e.g. [x, y], got {reference_points.shape[1]}."
        )
    if not isinstance(target_points, np.ndarray):
        raise ValueError(
            f"target_points should be numpy array, got {type(target_points)}."
        )
    if len(target_points.shape) != 2:
        raise ValueError(
            f"target_points should be of shape (N,2) meaning [...,[x, y],...], got {target_points.shape}."
        )
    if target_points.shape[1] != 2:
        raise ValueError(
            f"target_points 2nd dimension should be of size 2 e.g. [x, y], got {target_points.shape[1]}."
        )
    if len(reference_points) != len(target_points):
        raise ValueError(
            "The number of reference points and target points should be the same."
        )

    features = []

    for i in range(reference_points.shape[0]):
        reference_feature = create_feature_point(
            reference_points[i][0], reference_points[i][1], i, 0
        )
        target_feature = create_feature_point(
            target_points[i][0], target_points[i][1], i, 1
        )
        features.append(reference_feature)
        features.append(target_feature)

    features_msg = dai.TrackedFeatures()
    features_msg.trackedFeatures = features

    return features_msg



import depthai as dai

PRIMARY_COLOR = dai.Color(
    float(87 / 255), float(36 / 255), float(232 / 255), float(1.0)
)
TRANSPARENT_PRIMARY_COLOR = dai.Color(
    float(87 / 255), float(36 / 255), float(232 / 255), float(0.2)
)
SECONDARY_COLOR = dai.Color(
    float(240 / 255), float(240 / 255), float(240 / 255), float(1.0)
)



from typing import List

import depthai as dai

from depthai_nodes.node.base_host_node import BaseHostNode
from depthai_nodes.node.utils import nms_detections

from .tiling import Tiling


class TilesPatcher(BaseHostNode):
    """Handles the processing of tiled frames from neural network (NN) outputs, maps the
    detections from tiles back into the global frame, and sends out the combined
    detections for further processing.

    @ivar conf_thresh: Confidence threshold for filtering detections.
    @type conf_thresh: float
    @ivar iou_thresh: IOU threshold for non-max suppression.
    @type iou_thresh: float
    @ivar tile_manager: Manager responsible for handling tiling configurations.
    @type tile_manager: Tiling
    @ivar tile_buffer: Buffer to store tile detections temporarily.
    @type tile_buffer: list
    @ivar current_timestamp: Timestamp for the current frame being processed.
    @type current_timestamp: float
    @ivar expected_tiles_count: Number of tiles expected per frame.
    @type expected_tiles_count: int
    """

    def __init__(self) -> None:
        """Initializes the TilesPatcher node, sets default thresholds for confidence and
        IOU, and initializes buffers for tile processing."""
        super().__init__()
        self.name = "TilesPatcher"
        self.tile_manager = None
        self.conf_thresh = 0.3
        self.iou_thresh = 0.4

        self.tile_buffer = []
        self.current_timestamp = None
        self.expected_tiles_count = 0
        self._logger.debug("TilesPatcher initialized")

    def build(
        self, tile_manager: Tiling, nn: dai.Node.Output, conf_thresh=0.3, iou_thresh=0.4
    ) -> "TilesPatcher":
        """Configures the TilesPatcher node with the tile manager and links the neural
        network's output.

        @param tile_manager: The tiling manager responsible for tile positions and
            dimensions.
        @type tile_manager: Tiling
        @param nn: The output of the neural network node from which detections are
            received.
        @type nn: dai.Node.Output
        @param conf_thresh: Confidence threshold for filtering detections (default:
            0.3).
        @type conf_thresh: float
        @param iou_thresh: IOU threshold for non-max suppression (default: 0.4).
        @type iou_thresh: float
        @return: Returns self for method chaining.
        @rtype: TilesPatcher
        """
        self.conf_thresh = conf_thresh
        self.iou_thresh = iou_thresh
        self.tile_manager = tile_manager
        if (
            self.tile_manager.x is None
            or self.tile_manager.grid_size is None
            or self.tile_manager.overlap is None
        ):
            raise ValueError("Tile dimensions, grid size, or overlap not initialized.")
        self.expected_tiles_count = len(self.tile_manager.tile_positions)
        self.sendProcessingToPipeline(True)
        self.link_args(nn)
        self._logger.debug(
            f"TilesPatcher built with conf_thresh={conf_thresh}, iou_thresh={iou_thresh}"
        )
        return self

    def process(self, nn_output: dai.ImgDetections) -> None:
        """Processes each neural network output (detections) by mapping them from tiled
        patches back into the global frame and buffering them until all tiles for the
        current frame are processed.

        @param nn_output: The detections from the neural network's output.
        @type nn_output: dai.ImgDetections
        """
        self._logger.debug("Processing new input")
        timestamp = nn_output.getTimestamp()
        device_timestamp = nn_output.getTimestampDevice()

        if self.current_timestamp is None:
            self.current_timestamp = timestamp

        if self.current_timestamp != timestamp and len(self.tile_buffer) > 0:
            # new frame started, send the output for the previous frame
            self._send_output(self.current_timestamp, device_timestamp)
            self.tile_buffer = []

        self.current_timestamp = timestamp
        tile_index = nn_output.getSequenceNum()

        bboxes: List[dai.ImgDetection] = nn_output.detections
        mapped_bboxes = self._map_bboxes_to_global_frame(bboxes, tile_index)
        self.tile_buffer.append(mapped_bboxes)

        if len(self.tile_buffer) == self.expected_tiles_count:
            self._send_output(timestamp, device_timestamp)
            self.tile_buffer = []

    def _map_bboxes_to_global_frame(
        self, bboxes: List[dai.ImgDetection], tile_index: int
    ):
        """Maps bounding boxes from their local tile coordinates back to the global
        frame of the full image.

        @param bboxes: The bounding boxes to be mapped.
        @type bboxes: list[dai.ImgDetection]
        @param tile_index: The index of the tile being processed.
        @type tile_index: int
        @return: Mapped bounding boxes in the global image frame.
        @rtype: list[dai.ImgDetection]
        """
        tile_info = self._get_tile_info(tile_index)
        if (
            self.tile_manager is None
            or self.tile_manager.nn_shape is None
            or tile_info is None
        ):
            return []
        if tile_info is None:
            return []

        # Original tile coordinates in the global frame
        tile_x1, tile_y1, tile_x2, tile_y2 = tile_info["coords"]
        tile_actual_width = tile_x2 - tile_x1
        tile_actual_height = tile_y2 - tile_y1

        # Scaled dimensions (after resizing to fit NN input)
        scaled_width, scaled_height = tile_info["scaled_size"]
        nn_width, nn_height = self.tile_manager.nn_shape

        # Offsets due to padding
        x_offset = (nn_width - scaled_width) // 2
        y_offset = (nn_height - scaled_height) // 2

        # Scaling factors from scaled tile back to original tile dimensions
        scale_x = tile_actual_width / scaled_width
        scale_y = tile_actual_height / scaled_height

        global_bboxes = []
        for bbox in bboxes:
            # Convert bbox coordinates from normalized to NN input dimensions
            bbox_xmin_nn = bbox.xmin * nn_width
            bbox_ymin_nn = bbox.ymin * nn_height
            bbox_xmax_nn = bbox.xmax * nn_width
            bbox_ymax_nn = bbox.ymax * nn_height

            # Adjust for padding offsets to get coordinates in the scaled tile
            bbox_xmin_scaled = bbox_xmin_nn - x_offset
            bbox_ymin_scaled = bbox_ymin_nn - y_offset
            bbox_xmax_scaled = bbox_xmax_nn - x_offset
            bbox_ymax_scaled = bbox_ymax_nn - y_offset

            # Ensure coordinates are within the scaled tile dimensions
            bbox_xmin_scaled = max(0, min(bbox_xmin_scaled, scaled_width))
            bbox_ymin_scaled = max(0, min(bbox_ymin_scaled, scaled_height))
            bbox_xmax_scaled = max(0, min(bbox_xmax_scaled, scaled_width))
            bbox_ymax_scaled = max(0, min(bbox_ymax_scaled, scaled_height))

            # Map to original tile coordinates
            bbox_xmin_tile = bbox_xmin_scaled * scale_x
            bbox_ymin_tile = bbox_ymin_scaled * scale_y
            bbox_xmax_tile = bbox_xmax_scaled * scale_x
            bbox_ymax_tile = bbox_ymax_scaled * scale_y

            # Map to global image coordinates
            x1_global = tile_x1 + bbox_xmin_tile
            y1_global = tile_y1 + bbox_ymin_tile
            x2_global = tile_x1 + bbox_xmax_tile
            y2_global = tile_y1 + bbox_ymax_tile

            # Normalize global coordinates
            img_width, img_height = self.tile_manager.img_shape  # type: ignore
            normalized_bbox = dai.ImgDetection()
            normalized_bbox.label = bbox.label
            normalized_bbox.confidence = bbox.confidence
            normalized_bbox.xmin = x1_global / img_width
            normalized_bbox.ymin = y1_global / img_height
            normalized_bbox.xmax = x2_global / img_width
            normalized_bbox.ymax = y2_global / img_height

            global_bboxes.append(normalized_bbox)

        return global_bboxes

    def _get_tile_info(self, tile_index: int):
        """Retrieves the tile's coordinates and scaled dimensions based on the tile
        index.

        @param tile_index: The index of the tile.
        @type tile_index: int
        @return: Tile information for the given index.
        """
        if self.tile_manager is None or self.tile_manager.tile_positions is None:
            raise ValueError("Tile manager or tile positions not initialized.")
        if tile_index >= len(self.tile_manager.tile_positions):
            return None
        return self.tile_manager.tile_positions[tile_index]

    def _send_output(self, timestamp, device_timestamp):
        """Send the final combined bounding boxes as output when all tiles for a frame
        are processed.

        @param timestamp: The timestamp of the frame.
        @param device_timestamp: The timestamp of the frame on the device.
        """
        combined_bboxes: List[dai.ImgDetection] = []
        for bboxes in self.tile_buffer:
            combined_bboxes.extend(bboxes)

        if combined_bboxes:
            detection_list = nms_detections(
                combined_bboxes,
                conf_thresh=self.conf_thresh,
                iou_thresh=self.iou_thresh,
            )
        else:
            detection_list = []

        # Create ImgDetections message
        detections = dai.ImgDetections()
        detections.setTimestamp(timestamp)
        detections.setTimestampDevice(device_timestamp)
        detections.detections = detection_list

        self._logger.debug("Detections message created")

        self.out.send(detections)

        self._logger.debug("Message sent successfully")



import time
from typing import Callable, Dict, List, Optional, Union

import depthai as dai

from depthai_nodes.node.base_host_node import BaseHostNode

ProcessFnFrameOnlyType = Callable[["SnapsProducerFrameOnly", dai.ImgFrame], None]
ProcessFnType = Callable[["SnapsProducer", dai.ImgFrame, dai.Buffer], None]


class SnapsProducerFrameOnly(BaseHostNode):
    """A host node that helps with creating and sending snaps. If you also have
    additional message as input (e.g. detections) consider using `SnapsProducer` node
    instead.

    Attributes:
    ----------
    frame : dai.ImgFrame
        The input message for snap creation.
    time_interval : Union[int, float]
        Time interval between snaps if no custom process function is provided. Defaults to 60s.
    process_fn : Callable[['SnapsProducerFrameOnly', dai.ImgFrame], None]
        Custom processing function for business logic regarding snaps. If None defaults to sending only frames on time intervals.
    token : str
        Hub API token of the team you want to save snaps under. Can be also set through `DEPTHAI_HUB_API_KEY` env variable.
    url : str
        Custom URL for events service. By default, the URL is set to https://events-ingest.cloud.luxonis.com
    """

    def __init__(
        self,
        time_interval: Union[int, float] = 60.0,
        process_fn: Optional[ProcessFnFrameOnlyType] = None,
    ):
        super().__init__()
        self.last_update = time.time()

        self._em = dai.EventsManager()
        self._em.setLogResponse(True)
        self.setTimeInterval(time_interval)
        if process_fn is None:
            self._process_fn = None
        else:
            self.setProcessFn(process_fn)

        self._logger.debug(
            f"SnapsProducerFrameOnly initialized with time_interval={time_interval}"
        )

    def setToken(self, token: str):
        """Sets the Hub API token.

        @param token: Hub API token of your team.
        @type token: str
        """
        if not isinstance(token, str):
            raise ValueError("token must of of type string.")
        self._em.setToken(token)
        self._logger.debug(f"Token set to {token}")

    def setUrl(self, url: str):
        """Set the URL of the events service. By default, the URL is set to https://events-ingest.cloud.luxonis.com.

        @param url: URL of the events service.
        @type url: str
        """
        if not isinstance(url, str):
            raise ValueError("url must of of type string.")
        self._em.setUrl(url)
        self._logger.debug(f"Url set to {url}")

    def setTimeInterval(self, time_interval: Union[int, float]):
        """Sets time interval between snaps. Only relevant if using default processing.

        @param time_interval: Time interval between snaps for default sending in
            seconds.
        @type time_interval: Union[int, float]
        """
        if not (isinstance(time_interval, int) or isinstance(time_interval, float)):
            raise ValueError("time_interval must be of type int or float.")

        self.time_interval = time_interval
        self._logger.debug(f"Time interval set to {time_interval}")

    def setProcessFn(self, process_fn: ProcessFnFrameOnlyType):
        """Sets custom processing function.

        @param process_fn: Custom snaps processing function.
        @type process_fn: Callable[['SnapsProducerFrameOnly', dai.ImgFrame], None]
        """
        if not isinstance(process_fn, Callable):
            raise ValueError("process_fn must be a function.")

        self._process_fn = process_fn
        self._logger.debug("Process function set")

    def build(
        self,
        frame: dai.Node.Output,
        time_interval: Union[int, float] = 60.0,
        process_fn: Optional[ProcessFnFrameOnlyType] = None,
    ) -> "SnapsProducerFrameOnly":
        """Configures the node.

        @param frame: The input message for snap creation.
        @type frame: dai.Node.Output
        @param time_interval: Time interval between snaps for default sending in
            seconds. Defualts to 60.
        @type time_interval: Union[int, float]
        @param process_fn: Custom snaps processing function. Defaults to None.
        @type process_fn: Callable[['SnapsProducerFrameOnly', dai.ImgFrame], None]
        @return: The node object which handles snap creation and sending.
        @rtype: SnapsProducerFrameOnly
        """
        self.link_args(frame)
        self.setTimeInterval(time_interval)
        if process_fn is not None:
            self.setProcessFn(process_fn)
        self._logger.debug(
            f"SnapsProducerFrameOnly built with time_interval={time_interval}"
        )
        return self

    def process(self, frame: dai.Buffer):
        """Processes incoming frames and sends out snaps. If not custom process function
        is set then it sends only one frame every `time_interval` seconds. If using
        custom process function make sure to call `self.sendSnap()` at the end.

        @param frame: The input message for snap creation.
        @type frame: dai.ImgFrame
        """
        self._logger.debug("Processing new input")
        assert isinstance(frame, dai.ImgFrame)
        if self._process_fn is None:
            self.sendSnap("frame", frame)
        else:
            self._process_fn(self, frame)

    def sendSnap(
        self,
        name: str,
        frame: dai.ImgFrame,
        data: List[dai.EventData] = [],  # noqa: B006
        tags: List[str] = [],  # noqa: B006
        extra_data: Dict[str, str] = {},  # noqa: B006
        device_serial_num: str = "",
    ) -> bool:
        """Function that creates the snap and sends it out if time from last snap is
        greater than time_interval. Make sure to call this function from any custom
        processing functions to send snaps. Returns True if snap was sent.

        @param name: Name of the snap.
        @type name: str
        @param frame: Image frame to send.
        @type frame: dai.ImgFrame
        @param data: List of EventData objects to send. Defualts to [].
        @type data: List[dai.EventData]
        @param tags: List of tags to send. Defaults to [].
        @type tags: List[str]
        @param extra_data: Extra data to send. Defaults to {}.
        @type extra_data: Dict[str, str]
        @param device_serial_num: Device serial number. Defualts to ''.
        @type device_serial_num: str
        @return: True if snap was sent out else False.
        @rtype: bool
        """
        now = time.time()
        if now > self.last_update + self.time_interval:
            out = self._em.sendSnap(
                name=name,
                imgFrame=frame,
                data=data,
                tags=tags,
                extraData=extra_data,
                deviceSerialNo=device_serial_num,
            )
            if out:
                self._logger.info(f"Snap `{name}` sent")
                self.last_update = now
            return out
        return False


class SnapsProducer(SnapsProducerFrameOnly):
    """A host node that helps with creating and sending snaps. If you only have frame as
    input consider using `SnapsProducerFrameOnly` node instead.

    Attributes:
    ----------
    frame : dai.ImgFrame
        The frame input message for snap creation.
    msg : dai.Buffer
        The additional input message for snap creation.
    time_interval : float
        Time interval between snaps if no custom process function is provided. Defaults to 60s.
    process_fn : Callable[['SnapsProducer', dai.ImgFrame, dai.Buffer], None]
        Custom processing function for business logic regarding snaps. If None defaults to sending only frames on time intervals.
    token : str
        Hub API token of the team you want to save snaps under. Can be also set through `DEPTHAI_HUB_API_KEY` env variable.
    url : str
        Custom URL for events service. By default, the URL is set to https://events-ingest.cloud.luxonis.com
    """

    def setProcessFn(self, process_fn: ProcessFnType):
        """Sets custom processing function.

        @param process_fn: Custom snaps processing function.
        @type process_fn: Callable[['SnapsProducer', dai.ImgFrame, dai.Buffer], None]
        """
        if not isinstance(process_fn, Callable):
            raise ValueError("process_fn must be a function.")

        self._process_fn = process_fn
        self._logger.debug("Process function set")

    def build(
        self,
        frame: dai.Node.Output,
        msg: dai.Node.Output,
        time_interval: Union[int, float] = 60.0,
        process_fn: Optional[ProcessFnType] = None,
    ) -> "SnapsProducer":
        """Configures the node.

        @param frame: The frame input message for snap creation.
        @type frame: dai.Node.Output
        @param msg: The additonal input message for snap creation.
        @type msg: dai.Node.Output
        @param time_interval: Time interval between snaps for default sending in
            seconds. Defualts to 60.
        @type time_interval: Union[int, float]
        @param process_fn: Custom snaps processing function. Defaults to None.
        @type process_fn: Optional[Callable[['SnapsProducer', dai.ImgFrame, dai.Buffer],
            None]]
        @return: The node object which handles snap creation and sending.
        @rtype: SnapsProducer
        """
        self.link_args(frame, msg)
        self.setTimeInterval(time_interval)
        if process_fn is not None:
            self.setProcessFn(process_fn)
        self._logger.debug(f"SnapsProducer built with time_interval={time_interval}")
        return self

    def process(self, frame: dai.Buffer, msg: dai.Buffer):
        """Processes incoming frames and sends out snaps. If not custom process function
        is set then it sends only one frame every `time_interval` seconds. If using
        custom process function make sure to call `self.sendSnap()` at the end.

        @param frame: The frame input message for snap creation.
        @type frame: dai.ImgFrame
        @param frame: The additional input message for snap creation.
        @type frame: dai.Buffer
        """
        self._logger.debug("Processing new input")
        assert isinstance(frame, dai.ImgFrame)
        if self._process_fn is None:
            self.sendSnap("frame", frame)
        else:
            self._process_fn(self, frame, msg)



"""mediapipe.py.

Description: This script contains utility functions for decoding the output of the MediaPipe hand tracking model.

This script contains code that is based on or directly taken from a public GitHub repository:
https://github.com/geaxgx/depthai_hand_tracker

Original code author(s): geaxgx

License: MIT License

Copyright (c) [2021] [geax]
"""

import math
from collections import namedtuple

import numpy as np


class HandRegion:
    """
    Attributes:
    pd_score : detection score
    pd_box : detection box [x, y, w, h], normalized [0,1] in the squared image
    pd_kps : detection keypoints coordinates [x, y], normalized [0,1] in the squared image
    rect_x_center, rect_y_center : center coordinates of the rotated bounding rectangle, normalized [0,1] in the squared image
    rect_w, rect_h : width and height of the rotated bounding rectangle, normalized in the squared image (may be > 1)
    rotation : rotation angle of rotated bounding rectangle with y-axis in radian
    rect_x_center_a, rect_y_center_a : center coordinates of the rotated bounding rectangle, in pixels in the squared image
    rect_w, rect_h : width and height of the rotated bounding rectangle, in pixels in the squared image
    rect_points : list of the 4 points coordinates of the rotated bounding rectangle, in pixels
            expressed in the squared image during processing,
            expressed in the source rectangular image when returned to the user
    """

    def __init__(self, pd_score=None, pd_box=None, pd_kps=None):
        self.pd_score = pd_score  # Palm detection score
        self.pd_box = pd_box  # Palm detection box [x, y, w, h] normalized
        self.pd_kps = pd_kps  # Palm detection keypoints


SSDAnchorOptions = namedtuple(
    "SSDAnchorOptions",
    [
        "num_layers",
        "min_scale",
        "max_scale",
        "input_size_height",
        "input_size_width",
        "anchor_offset_x",
        "anchor_offset_y",
        "strides",
        "aspect_ratios",
        "reduce_boxes_in_lowest_layer",
        "interpolated_scale_aspect_ratio",
        "fixed_anchor_size",
    ],
)


def calculate_scale(min_scale, max_scale, stride_index, num_strides):
    if num_strides == 1:
        return (min_scale + max_scale) / 2
    else:
        return min_scale + (max_scale - min_scale) * stride_index / (num_strides - 1)


def generate_anchors(options):
    """
    option : SSDAnchorOptions
    # https://github.com/google/mediapipe/blob/master/mediapipe/calculators/tflite/ssd_anchors_calculator.cc
    """
    anchors = []
    layer_id = 0
    n_strides = len(options.strides)
    while layer_id < n_strides:
        anchor_height = []
        anchor_width = []
        aspect_ratios = []
        scales = []
        # For same strides, we merge the anchors in the same order.
        last_same_stride_layer = layer_id
        while (
            last_same_stride_layer < n_strides
            and options.strides[last_same_stride_layer] == options.strides[layer_id]
        ):
            scale = calculate_scale(
                options.min_scale, options.max_scale, last_same_stride_layer, n_strides
            )
            if last_same_stride_layer == 0 and options.reduce_boxes_in_lowest_layer:
                # For first layer, it can be specified to use predefined anchors.
                aspect_ratios += [1.0, 2.0, 0.5]
                scales += [0.1, scale, scale]
            else:
                aspect_ratios += options.aspect_ratios
                scales += [scale] * len(options.aspect_ratios)
                if options.interpolated_scale_aspect_ratio > 0:
                    if last_same_stride_layer == n_strides - 1:
                        scale_next = 1.0
                    else:
                        scale_next = calculate_scale(
                            options.min_scale,
                            options.max_scale,
                            last_same_stride_layer + 1,
                            n_strides,
                        )
                    scales.append(math.sqrt(scale * scale_next))
                    aspect_ratios.append(options.interpolated_scale_aspect_ratio)
            last_same_stride_layer += 1

        for i, r in enumerate(aspect_ratios):
            ratio_sqrts = math.sqrt(r)
            anchor_height.append(scales[i] / ratio_sqrts)
            anchor_width.append(scales[i] * ratio_sqrts)

        stride = options.strides[layer_id]
        feature_map_height = math.ceil(options.input_size_height / stride)
        feature_map_width = math.ceil(options.input_size_width / stride)

        for y in range(feature_map_height):
            for x in range(feature_map_width):
                for anchor_id in range(len(anchor_height)):
                    x_center = (x + options.anchor_offset_x) / feature_map_width
                    y_center = (y + options.anchor_offset_y) / feature_map_height
                    # new_anchor = Anchor(x_center=x_center, y_center=y_center)
                    if options.fixed_anchor_size:
                        new_anchor = [x_center, y_center, 1.0, 1.0]
                        # new_anchor.w = 1.0
                        # new_anchor.h = 1.0
                    else:
                        new_anchor = [
                            x_center,
                            y_center,
                            anchor_width[anchor_id],
                            anchor_height[anchor_id],
                        ]
                        # new_anchor.w = anchor_width[anchor_id]
                        # new_anchor.h = anchor_height[anchor_id]
                    anchors.append(new_anchor)

        layer_id = last_same_stride_layer
    return np.array(anchors)


def generate_handtracker_anchors(input_size_width, input_size_height):
    # https://github.com/google/mediapipe/blob/master/mediapipe/modules/palm_detection/palm_detection_cpu.pbtxt
    anchor_options = SSDAnchorOptions(
        num_layers=4,
        min_scale=0.1484375,
        max_scale=0.75,
        input_size_height=input_size_height,
        input_size_width=input_size_width,
        anchor_offset_x=0.5,
        anchor_offset_y=0.5,
        strides=[8, 16, 16, 16],
        aspect_ratios=[1.0],
        reduce_boxes_in_lowest_layer=False,
        interpolated_scale_aspect_ratio=1.0,
        fixed_anchor_size=True,
    )
    return generate_anchors(anchor_options)


def decode_bboxes(score_thresh, scores, bboxes, anchors, scale=128, best_only=False):
    # Wi, hi : NN input shape
    # mediapipe/calculators/tflite/tflite_tensors_to_detections_calculator.cc # Decodes
    # the detection tensors generated by the model, based on # the SSD anchors and the
    # specification in the options, into a vector of # detections. Each detection
    # describes a detected object.

    # https://github.com/google/mediapipe/blob/master/mediapipe/modules/palm_detection/palm_detection_cpu.pbtxt :
    # node {
    #     calculator: "TensorsToDetectionsCalculator"
    #     input_stream: "TENSORS:detection_tensors"
    #     input_side_packet: "ANCHORS:anchors"
    #     output_stream: "DETECTIONS:unfiltered_detections"
    #     options: {
    #         [mediapipe.TensorsToDetectionsCalculatorOptions.ext] {
    #         num_classes: 1
    #         num_boxes: 896
    #         num_coords: 18
    #         box_coord_offset: 0
    #         keypoint_coord_offset: 4
    #         num_keypoints: 7
    #         num_values_per_keypoint: 2
    #         sigmoid_score: true
    #         score_clipping_thresh: 100.0
    #         reverse_output_order: true

    #         x_scale: 128.0
    #         y_scale: 128.0
    #         h_scale: 128.0
    #         w_scale: 128.0
    #         min_score_thresh: 0.5
    #         }
    #     }
    # }
    # node {
    #     calculator: "TensorsToDetectionsCalculator"
    #     input_stream: "TENSORS:detection_tensors"
    #     input_side_packet: "ANCHORS:anchors"
    #     output_stream: "DETECTIONS:unfiltered_detections"
    #     options: {
    #         [mediapipe.TensorsToDetectionsCalculatorOptions.ext] {
    #         num_classes: 1
    #         num_boxes: 2016
    #         num_coords: 18
    #         box_coord_offset: 0
    #         keypoint_coord_offset: 4
    #         num_keypoints: 7
    #         num_values_per_keypoint: 2
    #         sigmoid_score: true
    #         score_clipping_thresh: 100.0
    #         reverse_output_order: true

    #         x_scale: 192.0
    #         y_scale: 192.0
    #         w_scale: 192.0
    #         h_scale: 192.0
    #         min_score_thresh: 0.5
    #         }
    #     }
    # }

    # scores: shape = [number of anchors 896 or 2016]
    # bboxes: shape = [ number of anchors x 18], 18 = 4 (bounding box : (cx,cy,w,h) + 14 (7 palm keypoints)

    regions = []
    scores = 1 / (1 + np.exp(-scores))
    if best_only:
        best_id = np.argmax(scores)
        if scores[best_id] < score_thresh:
            return regions
        det_scores = scores[best_id : best_id + 1]
        det_bboxes2 = bboxes[best_id : best_id + 1]
        det_anchors = anchors[best_id : best_id + 1]
    else:
        detection_mask = scores > score_thresh
        det_scores = scores[detection_mask]
        if det_scores.size == 0:
            return regions
    try:
        det_bboxes2 = bboxes[detection_mask]
        det_anchors = anchors[detection_mask]
    except Exception as e:
        raise IndexError(
            "Wrong parser scale set. Please use setScale method to set different scale according to model dimensions (e.g. 128)."
        ) from e

    det_bboxes = det_bboxes2 * np.tile(det_anchors[:, 2:4], 9) / scale + np.tile(
        det_anchors[:, 0:2], 9
    )
    det_bboxes[:, 2:4] = det_bboxes[:, 2:4] - det_anchors[:, 0:2]
    det_bboxes[:, 0:2] = det_bboxes[:, 0:2] - det_bboxes[:, 3:4] * 0.5

    for i in range(det_bboxes.shape[0]):
        score = det_scores[i]
        box = det_bboxes[i, 0:4]
        # Decoded detection boxes could have negative values for width/height due
        # to model prediction. Filter out those boxes
        if box[2] < 0 or box[3] < 0:
            continue
        kps = []
        # 0 : wrist
        # 1 : index finger joint
        # 2 : middle finger joint
        # 3 : ring finger joint
        # 4 : little finger joint
        # 5 :
        # 6 : thumb joint
        for kp in range(7):
            kps.append(det_bboxes[i, 4 + kp * 2 : 6 + kp * 2])
        regions.append(HandRegion(float(score), box, kps))
    return regions


def rect_transformation(regions, w, h, no_shift=False):
    """W, h : image input shape."""
    # https://github.com/google/mediapipe/blob/master/mediapipe/modules/hand_landmark/palm_detection_detection_to_roi.pbtxt
    # # Expands and shifts the rectangle that contains the palm so that it's likely
    # # to cover the entire hand.
    # node {
    # calculator: "RectTransformationCalculator"
    # input_stream: "NORM_RECT:raw_roi"
    # input_stream: "IMAGE_SIZE:image_size"
    # output_stream: "roi"
    # options: {
    #     [mediapipe.RectTransformationCalculatorOptions.ext] {
    #     scale_x: 2.6
    #     scale_y: 2.6
    #     shift_y: -0.5
    #     square_long: true
    #     }
    # }
    # IMHO 2.9 is better than 2.6. With 2.6, it may happen that finger tips stay outside of the bouding rotated rectangle
    scale_x = 2.9 if not no_shift else 1
    scale_y = 2.9 if not no_shift else 1
    shift_x = 0
    shift_y = -0.5 if not no_shift else 0
    for region in regions:
        width = region.rect_w
        height = region.rect_h
        rotation = 0
        if rotation == 0:
            region.rect_x_center_a = (region.rect_x_center + width * shift_x) * w
            region.rect_y_center_a = (region.rect_y_center + height * shift_y) * h
        else:
            x_shift = w * width * shift_x * math.cos(
                rotation
            ) - h * height * shift_y * math.sin(rotation)  # / w
            y_shift = w * width * shift_x * math.sin(
                rotation
            ) + h * height * shift_y * math.cos(rotation)  # / h
            region.rect_x_center_a = region.rect_x_center * w + x_shift
            region.rect_y_center_a = region.rect_y_center * h + y_shift

        long_side = max(width * w, height * h)
        region.rect_w_a = long_side * scale_x
        region.rect_h_a = long_side * scale_y
        region.rect_points = rotated_rect_to_points(
            region.rect_x_center_a,
            region.rect_y_center_a,
            region.rect_w_a,
            region.rect_h_a,
            region.rotation,
        )


def rotated_rect_to_points(cx, cy, w, h, rotation):
    b = math.cos(rotation) * 0.5
    a = math.sin(rotation) * 0.5
    p0x = cx - a * h - b * w
    p0y = cy + b * h - a * w
    p1x = cx + a * h - b * w
    p1y = cy - b * h - a * w
    p2x = int(2 * cx - p0x)
    p2y = int(2 * cy - p0y)
    p3x = int(2 * cx - p1x)
    p3y = int(2 * cy - p1y)
    p0x, p0y, p1x, p1y = int(p0x), int(p0y), int(p1x), int(p1y)
    return [[p0x, p0y], [p1x, p1y], [p2x, p2y], [p3x, p3y]]


def detections_to_rect(regions):
    # https://github.com/google/mediapipe/blob/master/mediapipe/modules/hand_landmark/palm_detection_detection_to_roi.pbtxt
    # # Converts results of palm detection into a rectangle (normalized by image size)
    # # that encloses the palm and is rotated such that the line connecting center of
    # # the wrist and MCP of the middle finger is aligned with the Y-axis of the
    # # rectangle.
    # node {
    #   calculator: "DetectionsToRectsCalculator"
    #   input_stream: "DETECTION:detection"
    #   input_stream: "IMAGE_SIZE:image_size"
    #   output_stream: "NORM_RECT:raw_roi"
    #   options: {
    #     [mediapipe.DetectionsToRectsCalculatorOptions.ext] {
    #       rotation_vector_start_keypoint_index: 0  # Center of wrist.
    #       rotation_vector_end_keypoint_index: 2  # MCP of middle finger.
    #       rotation_vector_target_angle_degrees: 90
    #     }
    #   }

    target_angle = math.pi * 0.5  # 90 = pi/2
    for region in regions:
        region.rect_w = region.pd_box[2]
        region.rect_h = region.pd_box[3]
        region.rect_x_center = region.pd_box[0] + region.rect_w / 2
        region.rect_y_center = region.pd_box[1] + region.rect_h / 2

        x0, y0 = region.pd_kps[0]  # wrist center
        x1, y1 = region.pd_kps[2]  # middle finger
        rotation = target_angle - math.atan2(-(y1 - y0), x1 - x0)
        region.rotation = normalize_radians(rotation)


def normalize_radians(angle):
    return angle - 2 * math.pi * math.floor((angle + math.pi) / (2 * math.pi))


def generate_anchors_and_decode(bboxes, scores, threshold=0.5, scale=192):
    """Generate anchors and decode bounding boxes for mediapipe hand detection model."""
    anchors = generate_handtracker_anchors(scale, scale)
    decoded_bboxes = decode_bboxes(threshold, scores, bboxes, anchors, scale=scale)
    detections_to_rect(decoded_bboxes)
    rect_transformation(decoded_bboxes, scale, scale, no_shift=True)
    return decoded_bboxes



from typing import List, Tuple

import depthai as dai

from .queue import OutputQueueMock


class OutputMock:
    """Output class to simulate the depthai output node."""

    def __init__(self):
        self._datatypes: List[Tuple[dai.DatatypeEnum, bool]] = []
        self._queues: List[OutputQueueMock] = []

    def setPossibleDatatypes(self, datatypes: List[Tuple[dai.DatatypeEnum, bool]]):
        self._datatypes = datatypes

    def getPossibleDatatypes(self) -> List[Tuple[dai.DatatypeEnum, bool]]:
        return self._datatypes

    def send(self, message):
        for queue in self._queues:
            queue.send(message)

    def createOutputQueue(
        self, checking_function=None, model_slug=None, parser_name=None
    ) -> OutputQueueMock:
        queue = OutputQueueMock(
            checking_function=checking_function,
            model_slug=model_slug,
            parser_name=parser_name,
        )
        self._queues.append(queue)
        return queue

    def link(self, node):
        pass

    def getParentPipeline(self):
        return None



from .arrays import create_img_frame, create_map, create_segmentation
from .classifications import create_classifications, create_classifications_sequence
from .clusters import create_clusters
from .detections import (
    create_img_detection,
    create_img_detection_extended,
    create_img_detections,
    create_img_detections_extended,
)
from .keypoints import create_keypoints
from .lines import create_lines
from .regression import create_regression

__all__ = [
    "create_img_frame",
    "create_map",
    "create_segmentation",
    "create_classifications",
    "create_classifications_sequence",
    "create_clusters",
    "create_img_detection",
    "create_img_detections",
    "create_img_detection_extended",
    "create_img_detections_extended",
    "create_keypoints",
    "create_lines",
    "create_regression",
]



from typing import Any, Dict


def decode_head(head) -> Dict[str, Any]:
    """Decode head object into a dictionary containing configuration details.

    @param head: The head object to decode.
    @type head: dai.nn_archive.v1.Head
    @return: A dictionary containing configuration details relevant to the head.
    @rtype: Dict[str, Any]
    """
    head_config = {}
    head_config["parser"] = head.parser
    head_config["outputs"] = head.outputs
    if head.metadata:
        head_config.update(head.metadata.extraParams)

    return head_config



from typing import Dict, List, Optional

import depthai as dai

from depthai_nodes.node.parsers import *
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils import decode_head


class ParserGenerator(dai.node.ThreadedHostNode):
    """General interface for instantiating parsers based on the provided model archive.

    The `build` method creates parsers based on the head information stored in the NN Archive. The method then returns a dictionary of these parsers.
    """

    DEVICE_PARSERS = ["YOLO", "SSD"]

    def build(
        self,
        nn_archive: dai.NNArchive,
        head_index: Optional[int] = None,
        host_only: bool = False,
    ) -> Dict:
        """Instantiates parsers based on the provided model archive.

        @param nn_archive: NN Archive of the model.
        @type nn_archive: dai.NNArchive
        @param head_index: Index of the head to be used for parsing. If not provided,
            each head will instantiate a separate parser.
        @type head_index: Optional[int]
        @param host_only: If True, only host parsers will be instantiated.
        @type host_only: bool
        @return: A dictionary of instantiated parsers.
        @rtype: Dict[int, BaseParser]
        """

        heads: List = nn_archive.getConfig().model.heads  # type: ignore

        indexes = range(len(heads))

        if len(heads) == 0:
            raise ValueError("No heads defined in the NN Archive.")

        if head_index:
            heads = [heads[head_index]]
            indexes = [head_index]

        parsers = {}
        pipeline = self.getParentPipeline()

        for index, head in zip(indexes, heads):
            parser_name = head.parser

            if parser_name in self.DEVICE_PARSERS:
                if host_only:
                    parser_name = self._getHostParserName(parser_name)
                else:
                    parser = pipeline.create(dai.node.DetectionParser)
                    parser.setNNArchive(nn_archive)
                    parsers[index] = parser
                    continue

            parser = globals().get(parser_name)

            if parser is None:
                raise ValueError(f"Parser {parser_name} not a valid parser class.")
            elif not issubclass(parser, BaseParser):
                raise ValueError(
                    f"Parser {parser_name} does not inherit from BaseParser class."
                )

            head_config = decode_head(head)
            head_config["model_inputs"] = []
            for input in nn_archive.getConfig().model.inputs:
                head_config["model_inputs"].append(
                    {"shape": input.shape, "layout": input.layout}
                )
            parsers[index] = pipeline.create(parser).build(head_config)

        return parsers

    def _getHostParserName(self, parser_name: str) -> str:
        if parser_name == "YOLO":
            return YOLOExtendedParser.__name__  # noqa: F405
        else:
            raise ValueError(
                f"Parser {parser_name} is not supported for host only mode."
            )

    def run(self):
        pass



import copy

import cv2
import depthai as dai
import numpy as np
from numpy.typing import NDArray


class Map2D(dai.Buffer):
    """Map2D class for storing a 2D map of floats.

    Attributes
    ----------
    map : NDArray[np.float32]
        2D map.
    width : int
        2D Map width.
    height : int
        2D Map height.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the Map2D object."""
        super().__init__()
        self._map: NDArray[np.float32] = np.array([])
        self._width: int = None
        self._height: int = None
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the Map2D class and copies the attributes.

        @return: A new instance of the Map2D class.
        @rtype: Map2D
        """
        new_obj = Map2D()
        new_obj.map = copy.deepcopy(self._map)
        new_obj.transformation = self._transformation
        return new_obj

    @property
    def map(self) -> NDArray[np.float32]:
        """Returns the 2D map.

        @return: 2D map.
        @rtype: NDArray[np.float32]
        """
        return self._map

    @map.setter
    def map(self, value: np.ndarray):
        """Sets the 2D map.

        @param value: 2D map.
        @type value: NDArray[np.float32]
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type float.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError(
                f"2D map must be of type np.ndarray, instead got {type(value)}."
            )
        if value.ndim != 2:
            raise ValueError("2D map must be a 2D array")
        if value.dtype != np.float32:
            raise ValueError("2D map must be an array of floats")
        self._map = value
        self._width = value.shape[1]
        self._height = value.shape[0]

    @property
    def width(self) -> int:
        """Returns the 2D map width.

        @return: 2D map width.
        @rtype: int
        """
        return self._width

    @property
    def height(self) -> int:
        """Returns the 2D map height.

        @return: 2D map height.
        @rtype: int
        """
        return self._height

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )

        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getVisualizationMessage(self) -> dai.ImgFrame:
        """Returns default visualization message for 2D maps in the form of a
        colormapped image."""
        img_frame = dai.ImgFrame()
        mask = self._map.copy()
        if np.any(mask < 1):
            mask = mask * 255
        mask = mask.astype(np.uint8)

        colored_mask = cv2.applyColorMap(mask, cv2.COLORMAP_PLASMA)
        return img_frame.setCvFrame(colored_mask, dai.ImgFrame.Type.BGR888i)



from .base_parser import BaseParser
from .classification import ClassificationParser
from .classification_sequence import ClassificationSequenceParser
from .detection import DetectionParser
from .embeddings import EmbeddingsParser
from .fastsam import FastSAMParser
from .hrnet import HRNetParser
from .image_output import ImageOutputParser
from .keypoints import KeypointParser
from .lane_detection import LaneDetectionParser
from .map_output import MapOutputParser
from .mediapipe_palm_detection import MPPalmDetectionParser
from .mlsd import MLSDParser
from .ppdet import PPTextDetectionParser
from .regression import RegressionParser
from .scrfd import SCRFDParser
from .segmentation import SegmentationParser
from .superanimal_landmarker import SuperAnimalParser
from .xfeat import XFeatMonoParser, XFeatStereoParser
from .yolo import YOLOExtendedParser
from .yunet import YuNetParser

__all__ = [
    "ImageOutputParser",
    "YuNetParser",
    "MPPalmDetectionParser",
    "SCRFDParser",
    "SegmentationParser",
    "SuperAnimalParser",
    "KeypointParser",
    "MLSDParser",
    "XFeatMonoParser",
    "XFeatStereoParser",
    "ClassificationParser",
    "YOLOExtendedParser",
    "FastSAMParser",
    "RegressionParser",
    "HRNetParser",
    "PPTextDetectionParser",
    "MapOutputParser",
    "ClassificationSequenceParser",
    "LaneDetectionParser",
    "BaseParser",
    "DetectionParser",
    "EmbeddingsParser",
]



from typing import List, Optional, Union

import numpy as np

from depthai_nodes import Classifications


def create_classification_message(
    classes: List[str], scores: Union[np.ndarray, List]
) -> Classifications:
    """Create a message for classification. The message contains the class names and
    their respective scores, sorted in descending order of scores.

    @param classes: A list containing class names.
    @type classes: List[str]
    @param scores: A numpy array of shape (n_classes,) containing the probability score of each class.
    @type scores: np.ndarray

    @return: A message with attributes `classes` and `scores`. `classes` is a list of classes, sorted in descending order of scores. `scores` is a list of the corresponding scores.
    @rtype: Classifications

    @raises ValueError: If the provided classes are None.
    @raises ValueError: If the provided classes are not a list.
    @raises ValueError: If the provided classes are empty.
    @raises ValueError: If the provided scores are None.
    @raises ValueError: If the provided scores are not a list or a numpy array.
    @raises ValueError: If the provided scores are empty.
    @raises ValueError: If the provided scores are not a 1D array.
    @raises ValueError: If the provided scores are not of type float.
    @raises ValueError: If the provided scores do not sum to 1.
    @raises ValueError: If the number of labels and scores mismatch.
    """
    if isinstance(classes, type(None)):
        raise ValueError("Classes should not be None.")

    if not isinstance(classes, list):
        raise ValueError(f"Classes should be a list, got {type(classes)}.")

    if len(classes) == 0:
        raise ValueError("Classes should not be empty.")

    if scores is None:
        raise ValueError("Scores should not be None.")

    if not isinstance(scores, np.ndarray) and not isinstance(scores, list):
        raise ValueError(
            f"Scores should be a list or a numpy array, got {type(scores)}."
        )

    if isinstance(scores, list):
        scores = np.array(scores)

    if len(scores) == 0:
        raise ValueError("Scores should not be empty.")

    if len(scores) != len(scores.flatten()):
        raise ValueError(f"Scores should be a 1D array, got {scores.shape}.")

    scores = scores.flatten()

    if not np.issubdtype(scores.dtype, np.floating):
        raise ValueError(f"Scores should be of type float, got {scores.dtype}.")

    scores = scores.astype(np.float32)

    if any([value < 0 or value > 1 for value in scores]):
        raise ValueError(
            f"Scores list must contain probabilities between 0 and 1, instead got {scores}."
        )

    if not np.isclose(np.sum(scores), 1.0, atol=1e-1):
        raise ValueError(f"Scores should sum to 1, got {np.sum(scores)}.")

    if len(scores) != len(classes):
        raise ValueError(
            f"Number of labels and scores mismatch. Provided {len(scores)} scores and {len(classes)} class names."
        )

    classification_msg = Classifications()
    sorted_args = np.argsort(-scores, kind="stable")
    scores = scores[sorted_args]

    classification_msg.classes = [classes[i] for i in sorted_args]
    classification_msg.scores = scores

    return classification_msg


def create_classification_sequence_message(
    classes: List[str],
    scores: Union[np.ndarray, List],
    ignored_indexes: Optional[List[int]] = None,
    remove_duplicates: bool = False,
    concatenate_classes: bool = False,
) -> Classifications:
    """Creates a message for a multi-class sequence. The message contains the class
    names and their respective scores, ordered according to the sequence. The 'scores'
    array is a sequence of probabilities for each class at each position in the
    sequence.

    @param classes: A list of class names, with length 'n_classes'.
    @type classes: List
    @param scores: A numpy array of shape (sequence_length, n_classes) containing the (row-wise) probability distributions over the classes.
    @type scores: np.ndarray
    @param ignored_indexes: A list of indexes to ignore during classification generation (e.g., background class, padding class). Defaults to None.
    @type ignored_indexes: Optional[List[int]]
    @param remove_duplicates: If True, removes consecutive duplicates from the sequence. Defaults to False.
    @type remove_duplicates: bool
    @param concatenate_classes: If True, concatenates consecutive classes based on the space character. Defaults to False.
    @type concatenate_classes: bool
    @return: A Classification message with attributes `classes` and `scores`, where `classes` is a list of class names and `scores` is a list of corresponding scores.
    @rtype: Classifications
    @raises ValueError: If 'classes' is not a list of strings.
    @raises ValueError: If 'scores' is not a 2D array of list of shape (sequence_length, n_classes).
    @raises ValueError: If the number of classes does not match the number of columns in 'scores'.
    @raises ValueError: If any score is not in the range [0, 1].
    @raises ValueError: If the probabilities in any row of 'scores' do not sum to 1.
    @raises ValueError: If 'ignored_indexes' in not None or a list of valid indexes within the range [0, n_classes - 1].
    """

    if not isinstance(classes, List):
        raise ValueError(f"Classes should be a list, got {type(classes)}.")

    if isinstance(scores, List):
        scores = np.array(scores)

    if len(scores.shape) != 2:
        raise ValueError(f"Scores should be a 2D array, got {scores.shape}.")

    if scores.shape[1] != len(classes):
        raise ValueError(
            f"Number of classes and scores mismatch. Provided {len(classes)} class names and {scores.shape[1]} scores."
        )

    if np.any(scores < 0) or np.any(scores > 1):
        raise ValueError("Scores should be in the range [0, 1].")

    if np.any(~np.isclose(scores.sum(axis=1), 1.0, atol=1e-2)):
        raise ValueError(
            f"Each row of scores should sum to 1, got {scores.sum(axis=1)}."
        )

    scores = scores.astype(np.float32)

    if ignored_indexes is not None:
        if not isinstance(ignored_indexes, List):
            raise ValueError(
                f"Ignored indexes should be a list, got {type(ignored_indexes)}."
            )
        if not all(isinstance(index, int) for index in ignored_indexes):
            raise ValueError("Ignored indexes should be integers.")
        if np.any(np.array(ignored_indexes) < 0) or np.any(
            np.array(ignored_indexes) >= len(classes)
        ):
            raise ValueError(
                "Ignored indexes should be integers in the range [0, num_classes -1]."
            )

    selection = np.ones(len(scores), dtype=bool)
    indexes = np.argmax(scores, axis=1)

    if remove_duplicates:
        selection[1:] = indexes[1:] != indexes[:-1]

    if ignored_indexes is not None:
        selection &= np.array([index not in ignored_indexes for index in indexes])

    class_list: List[str] = [classes[i] for i in indexes[selection]]
    score_list = np.max(scores, axis=1)[selection]

    if (
        concatenate_classes
        and len(class_list) > 1
        and all(len(word) <= 1 for word in class_list)
    ):
        concatenated_scores = []
        concatenated_words = "".join(class_list).split()
        cumsumlist = np.cumsum([len(word) for word in concatenated_words])

        start_index = 0
        for num_spaces, end_index in enumerate(cumsumlist):
            word_scores = score_list[start_index + num_spaces : end_index + num_spaces]
            concatenated_scores.append(np.mean(word_scores))
            start_index = end_index

        class_list = concatenated_words
        score_list = np.array(concatenated_scores)

    elif (
        concatenate_classes
        and len(class_list) > 1
        and any(len(word) >= 2 for word in class_list)
    ):
        class_list = [" ".join(class_list)]
        mean_score = np.mean(score_list)
        score_list = np.array([mean_score])

    classification_msg = Classifications()

    classification_msg.classes = class_list
    classification_msg.scores = score_list

    return classification_msg



from datetime import timedelta

import depthai as dai

from depthai_nodes.utils.annotation_helper import AnnotationHelper


def test_empty_build():
    annotation_helper = AnnotationHelper()
    annots = annotation_helper.build(timedelta(), 0)
    assert len(annots.annotations[0].points) == 0
    assert len(annots.annotations[0].circles) == 0
    assert len(annots.annotations[0].texts) == 0


def test_draw_line():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_line((0.0, 0.0), (1.0, 1.0), (1.0, 0.0, 0.0, 1.0), 2)
    annots = annotation_helper.build(timedelta(), 0)

    assert [(i.x, i.y) for i in annots.annotations[0].points[0].points] == [
        (0.0, 0.0),
        (1.0, 1.0),
    ]
    assert annots.annotations[0].points[0].thickness == 2
    assert annots.annotations[0].points[0].type == dai.PointsAnnotationType.LINE_STRIP
    assert annots.annotations[0].points[0].outlineColor.r == 1.0
    assert annots.annotations[0].points[0].outlineColor.g == 0.0
    assert annots.annotations[0].points[0].outlineColor.b == 0.0
    assert annots.annotations[0].points[0].outlineColor.a == 1.0


def test_draw_polyline():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_polyline(
        [(0.0, 0.0), (0.5, 0.5), (1.0, 1.0)],
        (1.0, 0.0, 0.0, 1.0),
        (0.0, 1.0, 0.0, 1.0),
        2,
        True,
    )
    annots = annotation_helper.build(timedelta(), 0)

    assert [(i.x, i.y) for i in annots.annotations[0].points[0].points] == [
        (0.0, 0.0),
        (0.5, 0.5),
        (1.0, 1.0),
    ]
    assert annots.annotations[0].points[0].thickness == 2
    assert annots.annotations[0].points[0].type == dai.PointsAnnotationType.LINE_LOOP
    assert annots.annotations[0].points[0].outlineColor.r == 1.0
    assert annots.annotations[0].points[0].outlineColor.g == 0.0
    assert annots.annotations[0].points[0].outlineColor.b == 0.0
    assert annots.annotations[0].points[0].outlineColor.a == 1.0
    assert annots.annotations[0].points[0].fillColor.r == 0.0
    assert annots.annotations[0].points[0].fillColor.g == 1.0
    assert annots.annotations[0].points[0].fillColor.b == 0.0
    assert annots.annotations[0].points[0].fillColor.a == 1.0


def test_draw_polyline_closed():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_polyline(
        [(0.0, 0.0), (0.5, 0.5), (1.0, 1.0)],
        (1.0, 0.0, 0.0, 1.0),
        None,
        2,
        False,
    )
    annots = annotation_helper.build(timedelta(), 0)

    assert [(i.x, i.y) for i in annots.annotations[0].points[0].points] == [
        (0.0, 0.0),
        (0.5, 0.5),
        (1.0, 1.0),
    ]
    assert annots.annotations[0].points[0].thickness == 2
    assert annots.annotations[0].points[0].type == dai.PointsAnnotationType.LINE_STRIP
    assert annots.annotations[0].points[0].outlineColor.r == 1.0
    assert annots.annotations[0].points[0].outlineColor.g == 0.0
    assert annots.annotations[0].points[0].outlineColor.b == 0.0
    assert annots.annotations[0].points[0].outlineColor.a == 1.0
    assert annots.annotations[0].points[0].fillColor.r == 0.0
    assert annots.annotations[0].points[0].fillColor.g == 0.0
    assert annots.annotations[0].points[0].fillColor.b == 0.0
    assert annots.annotations[0].points[0].fillColor.a == 0.0


def test_draw_circle():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_circle(
        (0.0, 0.0), 0.5, (1.0, 0.0, 0.0, 1.0), (0.0, 1.0, 0.0, 1.0), 2
    )
    annots = annotation_helper.build(timedelta(), 0)

    assert annots.annotations[0].circles[0].position.x == 0.0
    assert annots.annotations[0].circles[0].position.y == 0.0
    assert annots.annotations[0].circles[0].diameter == 1.0
    assert annots.annotations[0].circles[0].outlineColor.r == 1.0
    assert annots.annotations[0].circles[0].outlineColor.g == 0.0
    assert annots.annotations[0].circles[0].outlineColor.b == 0.0
    assert annots.annotations[0].circles[0].outlineColor.a == 1.0
    assert annots.annotations[0].circles[0].fillColor.r == 0.0
    assert annots.annotations[0].circles[0].fillColor.g == 1.0
    assert annots.annotations[0].circles[0].fillColor.b == 0.0
    assert annots.annotations[0].circles[0].fillColor.a == 1.0
    assert annots.annotations[0].circles[0].thickness == 2


def test_draw_rectangle():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_rectangle(
        (0.0, 0.0), (1.0, 1.0), (1.0, 0.0, 0.0, 1.0), (0.0, 1.0, 0.0, 1.0), 2
    )
    annots = annotation_helper.build(timedelta(), 0)

    assert [(i.x, i.y) for i in annots.annotations[0].points[0].points] == [
        (0.0, 0.0),
        (1.0, 0.0),
        (1.0, 1.0),
        (0.0, 1.0),
    ]
    assert annots.annotations[0].points[0].outlineColor.r == 1.0
    assert annots.annotations[0].points[0].outlineColor.g == 0.0
    assert annots.annotations[0].points[0].outlineColor.b == 0.0
    assert annots.annotations[0].points[0].outlineColor.a == 1.0
    assert annots.annotations[0].points[0].fillColor.r == 0.0
    assert annots.annotations[0].points[0].fillColor.g == 1.0
    assert annots.annotations[0].points[0].fillColor.b == 0.0
    assert annots.annotations[0].points[0].fillColor.a == 1.0
    assert annots.annotations[0].points[0].thickness == 2
    assert annots.annotations[0].points[0].type == dai.PointsAnnotationType.LINE_LOOP


def test_draw_text():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_text(
        "Test", (0.0, 0.0), (1.0, 0.0, 0.0, 1.0), (0.0, 1.0, 0.0, 1.0), 2
    )
    annots = annotation_helper.build(timedelta(), 0)

    assert annots.annotations[0].texts[0].text == "Test"
    assert annots.annotations[0].texts[0].position.x == 0.0
    assert annots.annotations[0].texts[0].position.y == 0.0
    assert annots.annotations[0].texts[0].backgroundColor.r == 0.0
    assert annots.annotations[0].texts[0].backgroundColor.g == 1.0
    assert annots.annotations[0].texts[0].backgroundColor.b == 0.0
    assert annots.annotations[0].texts[0].backgroundColor.a == 1.0
    assert annots.annotations[0].texts[0].textColor.r == 1.0
    assert annots.annotations[0].texts[0].textColor.g == 0.0
    assert annots.annotations[0].texts[0].textColor.b == 0.0
    assert annots.annotations[0].texts[0].textColor.a == 1.0
    assert annots.annotations[0].texts[0].fontSize == 2


def test_draw_points():
    annotation_helper = AnnotationHelper()
    annotation_helper.draw_points([(0.0, 0.0), (1.0, 1.0)], (1.0, 0.0, 0.0, 1.0), 2)
    annots = annotation_helper.build(timedelta(), 0)

    assert [(i.x, i.y) for i in annots.annotations[0].points[0].points] == [
        (0.0, 0.0),
        (1.0, 1.0),
    ]
    assert annots.annotations[0].points[0].thickness == 2
    assert annots.annotations[0].points[0].type == dai.PointsAnnotationType.POINTS
    assert annots.annotations[0].points[0].outlineColor.r == 1.0
    assert annots.annotations[0].points[0].outlineColor.g == 0.0
    assert annots.annotations[0].points[0].outlineColor.b == 0.0
    assert annots.annotations[0].points[0].outlineColor.a == 1.0



import numpy as np


def corners_to_rotated_bbox(corners: np.ndarray) -> np.ndarray:
    """Converts the corners of a bounding box to a rotated bounding box.

    @param corners: The corners of the bounding box. The corners are expected to be
        ordered by top-left, top-right, bottom-right, bottom-left.
    @type corners: np.ndarray
    @return: The rotated bounding box defined as [x_center, y_center, width, height,
        angle].
    @rtype: np.ndarray
    """
    if not isinstance(corners, np.ndarray):
        raise ValueError("Corners must be a numpy array.")
    if corners.shape != (4, 2):
        raise ValueError("Corners must be of shape (4, 2).")

    if corners[0][0] >= corners[1][0]:
        raise ValueError(
            "Corners must be ordered by top-left, top-right, bottom-right, bottom-left. Got top-right corner before top-left."
        )
    if corners[0][1] >= corners[3][1]:
        raise ValueError(
            "Corners must be ordered by top-left, top-right, bottom-right, bottom-left. Got bottom-left corner before top-left."
        )
    if corners[1][1] >= corners[2][1]:
        raise ValueError(
            "Corners must be ordered by top-left, top-right, bottom-right, bottom-left. Got bottom-right corner before top-right."
        )
    if corners[3][0] >= corners[2][0]:
        raise ValueError(
            "Corners must be ordered by top-left, top-right, bottom-right, bottom-left. Got bottom-left corner before bottom-right."
        )

    x_dist = corners[1][0] - corners[0][0]
    y_dist = corners[1][1] - corners[0][1]

    angle = np.degrees(np.arctan2(y_dist, x_dist))
    x_center, y_center = np.mean(corners, axis=0)
    width = np.linalg.norm(corners[0] - corners[1])
    height = np.linalg.norm(corners[1] - corners[2])

    return np.array([x_center, y_center, width, height, angle])


def rotated_bbox_to_corners(
    cx: float, cy: float, w: float, h: float, rotation: float
) -> np.ndarray:
    """Converts a rotated bounding box to the corners of the bounding box.

    @param cx: The x-coordinate of the center of the bounding box.
    @type cx: float
    @param cy: The y-coordinate of the center of the bounding box.
    @type cy: float
    @param w: The width of the bounding box.
    @type w: float
    @param h: The height of the bounding box.
    @type h: float
    @param rotation: The angle of the bounding box given in degrees.
    @type rotation: float
    @return: The corners of the bounding box.
    @rtype: np.ndarray
    """
    if not all(isinstance(i, float) for i in [cx, cy, w, h, rotation]):
        raise ValueError("All inputs must be floats.")
    if not -360 <= rotation <= 360:
        raise ValueError("Rotation must be between -360 and 360.")

    b = np.cos(rotation) * 0.5
    a = np.sin(rotation) * 0.5
    p0x = cx - a * h - b * w
    p0y = cy + b * h - a * w
    p1x = cx + a * h - b * w
    p1y = cy - b * h - a * w
    p2x = int(2 * cx - p0x)
    p2y = int(2 * cy - p0y)
    p3x = int(2 * cx - p1x)
    p3y = int(2 * cy - p1y)
    p0x, p0y, p1x, p1y = int(p0x), int(p0y), int(p1x), int(p1y)

    return np.array([[p0x, p0y], [p1x, p1y], [p2x, p2y], [p3x, p3y]])


def xyxy_to_xywh(bboxes: np.ndarray) -> np.ndarray:
    """Converts bounding boxes from [x_min, y_min, x_max, y_max] to [x_center, y_center,
    width, height].

    @param bboxes: The bounding boxes to convert.
    @type bboxes: np.ndarray
    @return: The converted bounding boxes.
    @rtype: np.ndarray
    """
    if not isinstance(bboxes, np.ndarray):
        raise ValueError("Bounding boxes must be a numpy array.")
    if len(bboxes.shape) != 2:
        raise ValueError("Bounding boxes must be of shape (N, 4).")

    if bboxes.shape[1] != 4:
        raise ValueError("Bounding boxes must be of shape (N, 4).")

    if not np.all(bboxes[:, :2] <= bboxes[:, 2:]):
        raise ValueError(
            "Bounding box coordinates must be in the format [x_min, y_min, x_max, y_max]."
        )

    x_center = (bboxes[:, 0] + bboxes[:, 2]) / 2
    y_center = (bboxes[:, 1] + bboxes[:, 3]) / 2
    width = bboxes[:, 2] - bboxes[:, 0]
    height = bboxes[:, 3] - bboxes[:, 1]

    return np.stack([x_center, y_center, width, height], axis=-1)


def xywh_to_xyxy(bboxes: np.ndarray) -> np.ndarray:
    """Convert bounding box coordinates from (x_center, y_center, width, height) to
    (x_min, y_min, x_max, y_max).

    @param bboxes: A numpy array of shape (N, 4) containing the bounding boxes in (x, y, width, height) format.
    @type np.ndarray
    @return: A numpy array of shape (N, 4) containing the bounding boxes in (x_min, y_min, x_max, y_max) format.
    @type np.ndarray
    """
    if not isinstance(bboxes, np.ndarray):
        raise ValueError("Bounding boxes must be a numpy array.")
    if len(bboxes.shape) != 2:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Got shape {bboxes.shape}."
        )
    if bboxes.shape[1] != 4:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Got shape {bboxes.shape}."
        )

    xyxy_bboxes = np.zeros_like(bboxes)
    xyxy_bboxes[:, 0] = bboxes[:, 0] - bboxes[:, 2] / 2  # x_min = x - w/2
    xyxy_bboxes[:, 1] = bboxes[:, 1] - bboxes[:, 3] / 2  # y_min = y - h/2
    xyxy_bboxes[:, 2] = bboxes[:, 0] + bboxes[:, 2] / 2  # x_max = x + w/2
    xyxy_bboxes[:, 3] = bboxes[:, 1] + bboxes[:, 3] / 2  # y_max = y + h/2

    return xyxy_bboxes


def top_left_wh_to_xywh(bboxes: np.ndarray) -> np.ndarray:
    """Converts bounding boxes from [top_left_x, top_left_y, width, height] to
    [x_center, y_center, width, height].

    @param bboxes: The bounding boxes to convert.
    @type bboxes: np.ndarray
    @return: The converted bounding boxes.
    @rtype: np.ndarray
    """
    if not isinstance(bboxes, np.ndarray):
        raise ValueError("Bounding boxes must be a numpy array.")
    if len(bboxes.shape) != 2:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Got shape {bboxes.shape}."
        )
    if bboxes.shape[1] != 4:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Got shape {bboxes.shape}."
        )

    x_center = bboxes[:, 0] + bboxes[:, 2] / 2
    y_center = bboxes[:, 1] + bboxes[:, 3] / 2
    width = bboxes[:, 2]
    height = bboxes[:, 3]

    return np.stack([x_center, y_center, width, height], axis=-1)


def normalize_bboxes(bboxes: np.ndarray, height: int, width: int) -> np.ndarray:
    """Normalize bounding box coordinates to (0, 1).

    @param bboxes: A numpy array of shape (N, 4) containing the bounding boxes.
    @type np.ndarray
    @param height: The height of the image.
    @type height: int
    @param width: The width of the image.
    @type width: int
    @return: A numpy array of shape (N, 4) containing the normalized bounding boxes.
    @type np.ndarray
    """
    if not isinstance(bboxes, np.ndarray):
        raise ValueError("Bounding boxes must be a numpy array.")
    if len(bboxes.shape) != 2:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Got shape {bboxes.shape}."
        )
    if bboxes.shape[1] != 4:
        raise ValueError(
            "Bounding boxes must be of shape (N, 4). Other options are currently not supported"
        )
    bboxes[:, 0] = bboxes[:, 0] / width
    bboxes[:, 1] = bboxes[:, 1] / height
    bboxes[:, 2] = bboxes[:, 2] / width
    bboxes[:, 3] = bboxes[:, 3] / height

    return bboxes
